import sys
import numpy as np
from scipy import ndimage
import warnings

import pdb

from scipy.optimize import fmin_slsqp
from scipy.interpolate import interp1d

import logging
LOGGER_NAME = 'FDTDStaggered'
logger = logging.getLogger(LOGGER_NAME)

try:
    from .StaggeredFDTD_2D_With_Relaxation_CUDA import StaggeredFDTD_2D_CUDA
    print ("StaggeredFDTD_2D_CUDA loaded")
except:
    print ("StaggeredFDTD_2D_CUDA NOT loaded")
try:
    from .StaggeredFDTD_2D_With_Relaxation_OPENCL import StaggeredFDTD_2D_OPENCL
    print ("StaggeredFDTD_2D_OPENCL loaded")
except:
     print ("StaggeredFDTD_2D_OPENCL NOT loaded")
try:
    from .StaggeredFDTD_2D_With_Relaxation_METAL import StaggeredFDTD_2D_METAL
    print ("StaggeredFDTD_2D_METAL loaded")
except:
    print ("StaggeredFDTD_2D_METAL NOT loaded")

#############################################################################################################
#This global dictionary specifies the order in the MaterialProperties array (Nx5) where N is the numbe of materials
StaggeredConstants={}
StaggeredConstants['ColDensity']=0
StaggeredConstants['ColLongSOS']=1
StaggeredConstants['ColShearSOS']=2
StaggeredConstants['ColLongAtt']=3
StaggeredConstants['ColShearAtt']=4

class PropagationModel2D:
    def StaggeredFDTD_2D_with_relaxation(self,MaterialMap,
                                         MaterialProperties,
                                         Frequency,
                                         SourceMap,
                                         SourceFunctions,
                                         SpatialStep,
                                         DurationSimulation,
                                         SensorMap,
                                         Ox=np.array([1]),
                                         Oy=np.array([1]),
                                         AlphaCFL=0.99,
                                         NDelta=12,
                                         ReflectionLimit=1.0000e-05,
                                         IntervalSnapshots=-1,
                                         COMPUTING_BACKEND=1,
                                         USE_SINGLE=True,
                                         SPP_ZONES=1,
                                         SPP_VolumeFraction=None,
                                         DT=None,
                                         QfactorCorrection=True,
                                         CheckOnlyParams=False,
                                         TypeSource=0,
                                         SelRMSorPeak=1,
                                         SelMapsRMSPeakList=['Pressure'],
                                         SelMapsSensorsList=['Vx','Vy'],
                                         SensorSubSampling=2,
                                         SensorStart=0,
                                         DefaultGPUDeviceName='TITAN',
                                         DefaultGPUDeviceNumber=0,
                                         SILENT=0,
                                         QCorrection=1.0,
                                         ManualGroupSize=np.array([-1,-1,-1]).astype(np.int32),
                                         ManualLocalSize=np.array([-1,-1,-1]).astype(np.int32)):
        '''
        Samuel Pichardo, Ph.D.
        2020

        Implementation of the 2D Sttagered-grid  Virieux scheme for the elastodynamic equation
        to calculate wave propagation in a heterogenous medium with liquid and solids
        (Virieux, Jean. "P-SV wave propagation in heterogeneous media; velocity-stress
        finite-difference method." Geophysics 51.4 (1986): 889-901.))
        The staggered grid method has the huge advantage of modeling correctly
        the propagation from liquid to solid and solid to liquid
        without the need of complicated coupled systems between acoustic and
        elastodynamic equations.

        We assume isotropic conditions (c11=c22, c12= Lamba Lame coeff, c66= Miu Lame coeff), with
        Cartesian square coordinate system.

        WE USE SI convention! meters, seconds, meters/s, Pa,kg/m^3, etc, for units.
        Version 4.0. We use now the rotated staggered grid to better model liquid-solids interfaces
        Saenger, Erik H., Norbert Gold, and Serge A. Shapiro.
        "Modeling the propagation of elastic waves using a modified finite-difference grid."
        Wave motion 31.1 (2000): 77-92.

        Version 3.0. We add true viscosity handling to account attenuation losses, following:
        Blanch, Joakim O., Johan OA Robertsson, and William W. Symes. "Modeling of a constant Q: Methodology and algorithm for an efficient and optimally inexpensive viscoelastic technique." Geophysics 60.1 (1995): 176-184.
        and
        Bohlen, Thomas. "Parallel 3-D viscoelastic finite difference seismic modelling." Computers & Geosciences 28.8 (2002): 887-899.

        version 2.0 April 2013.
        "One code to rule them all, one code to find them, one code to bring them all and in the darkness bind them". Sorry , I couln't avoid it :)
        The low-level function is based on a single code that produce  Python and Matlab modules, single or double precision, and to run either X64 or CUDA code. This will help a lot to propagate any important change in all
        implementations


        version 1.0 March 2013.
        The Virieux method is relatively easy to implement, the tricky point was
        to implement correctly, with the lowest memory footprint possible,
        the Perfect Matching  Layer (PML) method. I used the method of Collino and
        Tsogka "Application of the perfectly matched absorbing layer model to
        the linear elastodynamic problem in anisotropic
        heterogenous media" Geophysics, 66(1) pp 294-307. 2001.
        Both methods (Vireux + Collino&Tsogka) are quite referenced and validated.

        Use
        [Sensor,LastMap,SnapShots]=StaggeredFDTD_2D(MaterialMap,MaterialProperties,
        SourceMap,SourceFunctions,SpatialStep,DurationSimulation,SensorMap)

        where
        'MaterialMap' is an uint32 N1xN2 2D matrix where each different
        material is labeled. Start labeling from '0'
        'MaterialProperties' is a Nx3 matrix where N is the maximum number of
        materials in MaterialMap. Since the labeling should starts in 0, N must be max(MaterialMap(:))+1
        Column 1 in MaterialProperties refers to the density of the material,
        Column 2 is the compressional speed of sound
        Column 3 is the shear speed of sound

        'SourceMap' is a uint32 N1xN2 matrix where the presence of source is
        indicate. Any value different to 0 is considered as a source. The value
        indicate which source function must be refered.

        'SourceFunctions' is a N-sources x NumberTimeSteps matrix
        where N-sources is the maximun value of SourceMap. Each row in
        SourceFuntions is the evolution on time of the stress (xx,yy) (negative
        version of the pressure)


        You can specifu optional parameters 'IntervalSnapshots', 'ReflectionLimit' (for
        PML), 'NDelta. (also for  PML), 'AlphaCFL' (for the Courant convergence criteria)
        For the default values for the PML, reead below in the code and be
        cautious when modifying them, since your simulation mat get a lot of
        dispersion or reflections.

        '''


        SzMap=MaterialMap.shape
        # if TypeSource<2:
        if not(np.all(SzMap[0:2]==SourceMap.shape)  and np.all(SzMap[0:2]==SensorMap.shape)
                and np.all(SzMap[0:2]==Ox.shape) and np.all(SzMap[0:2]==Oy.shape) ):
                raise ValueError('The size SourceMap, Ox, Oy, MaterialMap, SensorMap must be equal!!!')
        # else:
        #     if not(np.all(SzMap[0:2]==SourceMap.shape)  and np.all(SzMap[0:2]==SensorMap.shape)):
        #         raise ValueError('The size SourceMap, MaterialMap, SensorMap must be equal!!!')
        #     if Ox.ndim==1:
        #         if not(Ox.size==1 and Oy.size==1  and Ox[0]==1 and Oy[0]==1 ):
        #             raise ValueError('When specifying a source for stress, Oy must remain equal to [1], and Ox can be either [1] or same dimensions as SourceMap')
        #         Ox=np.ones(MaterialMap.shape)
        #     else:
        #         if not( np.all(SzMap[0:2]==Ox.shape) and Oy.size==1  and Oy[0]==1 ):
        #             raise ValueError('When specifying a source for stress, Oy must remain equal to [1], and Ox can be either [1] or same dimensions as SourceMap')



        N1=SzMap[0]
        N2=SzMap[1]

        h=SpatialStep

        VMaxLong=np.max(MaterialProperties[:,np.int32(StaggeredConstants['ColLongSOS'])])

        ###%%%%%%


        dt,RhoMat,MiuMat, LambdaMiuMat, LambdaMat,TauLong,TauShear,TauSigma,AnalysisQFactorLong,AnalysisQFactorShear=self.CalculateMatricesForPropagation(MaterialMap,MaterialProperties,Frequency,QfactorCorrection,h,AlphaCFL,QCorrection=QCorrection)

        Omega=Frequency*2*np.pi

        PoisonRatio=LambdaMat.flatten()/(2.0*(LambdaMat.flatten()+MiuMat.flatten()))

        bValidPoisonRatio=True
        if np.max(PoisonRatio)>0.5:
            bValidPoisonRatio=False
            if CheckOnlyParams==False:
                raise ValueError('Poison ratio larger than 0.5!!!! are you sure of the values of speed of sound and density??')
        if np.min(PoisonRatio)<-1.0:
            bValidPoisonRatio=False
            if CheckOnlyParams==False:
                raise ValueError('Poison ratio smaller than -1!!!! are you sure of the values of speed of sound and density??')

        if dt < 0.0:
            raise ValueError('Invalid dt conditions!!! dt =' + str(dt))
        if CheckOnlyParams:
            return bValidPoisonRatio, dt

        if SPP_ZONES>1 and SPP_VolumeFraction is None:
            raise ValueError('SPP_VolumeFraction must contain a (N1,N2) matrix with the fraction of solid material')
        if SPP_ZONES>1:
            if np.all(SzMap[0:2]==SPP_VolumeFraction.shape)  is None:
                raise ValueError('SPP_VolumeFraction must contain a matrix with the fraction of solid material of same dimensions as material map')
        if SPP_ZONES>1 and (np.any(SPP_VolumeFraction<0) or np.any(SPP_VolumeFraction)>1.0):
            raise ValueError('SPP_VolumeFraction must contain values between 0.0 and 1.0 ')

        OneOverTauSigma=1.0/TauSigma

        MaterialMap2D=np.zeros((N1+1,N2+1),MaterialMap.dtype)
        MaterialMap2D[0:N1,0:N2]=MaterialMap
        MaterialMap2D[-1,:]=MaterialMap2D[-2,:]
        MaterialMap2D[:,-1]=MaterialMap2D[:,-2]


        #TODO: Dec 30, 2020: We have some issues with the PML if it is a solid....
        MaterialMap2D[0:NDelta,:]=0
        MaterialMap2D[N1-NDelta:,:]=0
        MaterialMap2D[:,0:NDelta]=0
        MaterialMap2D[:,N2-NDelta:]=0
 

        if DT!=None:
             if DT >dt:
                import warnings
                raise ValueError('Staggered:DT_INVALID The specified manual step is larger than the minimal optimal size, there is a risk of unstable calculation ' + str(DT) + ' ' +str(dt))

             else:
                print ('The specified manual step  is smaller than the minimal optimal size, calculations may take longer than required\n', DT,dt)
                dt=DT


        TimeVector=np.arange(0.0,DurationSimulation,dt)

        if SourceFunctions.shape[0]<np.max(SourceMap.flatten()):
             raise ValueError('The maximum identifier in SourceMap  is larger than the maximum source function (maximum row) in SourceFunctions')
        
        NumberSensorSubSampling=np.floor(TimeVector.size/SensorSubSampling)
        if (SensorStart <0) or (SensorStart>=NumberSensorSubSampling):
            raise ValueError('Invalid SensorStart value, it must be larger or equal to 0, and less than the size of the sensor length %i ' %(NumberSensorSubSampling))


        LengthSource = SourceFunctions.shape[1]

        delta=(NDelta-2.0)*h;# %% We do a trick to force the PML to act in the regions where there are calculations.
        delta=(NDelta)*h;# %% We do a trick to force the PML to act in the regions where there are calculations.

        Dx = np.zeros((NDelta+1,1))
        Dxhp = np.zeros((NDelta+1,1))

        d0 = np.log(1.0/ReflectionLimit)*3*VMaxLong/2/delta
        #%d0 = 80*8.63e-2*max(VLong)/delta;
        ddx=(d0*(np.arange(1,NDelta+1)*h/delta)**2)
        ddxhp=(d0*((np.arange(1,NDelta+1)-0.5)*h/delta)**2)

        Dx[0:NDelta,0]=np.flipud(ddx).flatten()
        Dxhp[0:NDelta,0]=np.flipud(ddxhp).flatten()

        InvDXDTplus=1.0/(1.0/dt + Dx/2)
        DXDTminus=(1.0/dt - Dx/2)
        InvDXDTplushp=1.0/(1.0/dt + Dxhp/2)
        DXDTminushp=(1.0/dt - Dxhp/2)
        #%%%%%%%%%%%%%%%% Perfect Matched Layer


        LambdaMiuMatOverH=LambdaMiuMat /h
        LambdaMatOverH=LambdaMat /h
        MiuMatOverH=MiuMat/h

        InvRhoMatH =1.0/RhoMat/h

        #%% the sensors are easy, just pass the indexes that need to be observed
        IndexSensors=np.nonzero(np.transpose(SensorMap).flatten()>0)[0]+1 #KEEP the +1, since in the low level function the index is substracted

        SnapshotsPos=[]
        SnapShots=[]
        if IntervalSnapshots>0:
            tPlot=1
            for n in range(TimeVector.size):
                if np.floor(TimeVector[n]/IntervalSnapshots)==tPlot:
                    SnapshotsPos.append(n)
                    SnapShots.append({'time':TimeVector[n]})
                    tPlot+=+1
        InputParam={}

        #We decode what maps to collect for RMS-Peak
        SelMapsRMSPeak=int(0)
        SelMapsSensors=int(0)
        curIndexRMS=0
        curIndexSensors=0
        IndexRMSMaps={}
        IndexSensorMaps={}
        curMask=int(0x0001)
        #Do not modify the order of this search without matching the low level functions!
        for pMap in ['Vx','Vy','Sigmaxx','Sigmayy','Sigmaxy','Pressure','Pressure_gx','Pressure_gy']:
            if pMap not in ['Pressure_gx','Pressure_gy']:
                if pMap in  SelMapsRMSPeakList:
                    SelMapsRMSPeak=SelMapsRMSPeak | curMask
                    IndexRMSMaps[pMap]=curIndexRMS
                    curIndexRMS+=1
                else:
                    IndexRMSMaps[pMap]=-1

            if pMap in  SelMapsSensorsList:
                SelMapsSensors=SelMapsSensors | curMask
                IndexSensorMaps[pMap]=curIndexSensors
                curIndexSensors+=1
            else:
                IndexSensorMaps[pMap]=-1
            curMask=curMask<<1


        if USE_SINGLE==False:
            InputParam['InvDXDTplus']=InvDXDTplus;
            InputParam['DXDTminus']=DXDTminus;
            InputParam['InvDXDTplushp']=InvDXDTplushp;
            InputParam['DXDTminushp']=DXDTminushp;
            InputParam['LambdaMiuMatOverH']=LambdaMiuMatOverH;
            InputParam['LambdaMatOverH']=LambdaMatOverH;
            InputParam['MiuMatOverH']=MiuMatOverH;
            InputParam['TauLong']=TauLong;
            InputParam['TauShear']=TauShear;
            InputParam['OneOverTauSigma']=OneOverTauSigma;
            InputParam['InvRhoMatH']=InvRhoMatH;
            InputParam['SourceFunctions']=SourceFunctions;
            InputParam['DT']=dt;
            InputParam['Ox']=Ox;
            InputParam['Oy']=Oy;
        else:
            InputParam['InvDXDTplus']=InvDXDTplus.astype(np.float32)
            InputParam['DXDTminus']=DXDTminus.astype(np.float32)
            InputParam['InvDXDTplushp']=InvDXDTplushp.astype(np.float32)
            InputParam['DXDTminushp']=DXDTminushp.astype(np.float32)
            InputParam['LambdaMiuMatOverH']=LambdaMiuMatOverH.astype(np.float32)
            InputParam['LambdaMatOverH']=LambdaMatOverH.astype(np.float32)
            InputParam['MiuMatOverH']=MiuMatOverH.astype(np.float32)
            InputParam['TauLong']=TauLong.astype(np.float32)
            InputParam['TauShear']=TauShear.astype(np.float32)
            InputParam['OneOverTauSigma']=OneOverTauSigma.astype(np.float32)
            InputParam['InvRhoMatH']=InvRhoMatH.astype(np.float32)
            InputParam['SourceFunctions']=SourceFunctions.astype(np.float32)
            InputParam['DT']=np.float32(dt);
            InputParam['Ox']=Ox.astype(np.float32)
            InputParam['Oy']=Oy.astype(np.float32)


        InputParam['MaterialMap']=MaterialMap2D;
        InputParam['IndexSensorMap']=np.uint32(IndexSensors)
        InputParam['N1']=np.uint32(N1)
        InputParam['N2']=np.uint32(N2)
        InputParam['SILENT']=np.uint32(SILENT)
        InputParam['TypeSource']=np.uint32(TypeSource)
        InputParam['TimeSteps']=np.uint32(TimeVector.size)
        InputParam['SourceMap']=np.uint32(SourceMap)
        InputParam['SnapshotsPos']=np.uint32(SnapshotsPos)
        InputParam['PMLThickness']=np.uint32(NDelta)
        InputParam['SelRMSorPeak']=np.uint32(SelRMSorPeak)
        InputParam['SelMapsRMSPeak']=np.uint32(SelMapsRMSPeak)
        InputParam['SensorSubSampling']=np.uint32(SensorSubSampling)
        InputParam['SensorStart']=np.uint32(SensorStart)
        InputParam['SelMapsSensors']=np.uint32(SelMapsSensors)
        InputParam['LengthSource']=np.uint32(LengthSource); #%we need now to provided a limit how much the source lasts
        InputParam['DefaultGPUDeviceName']=DefaultGPUDeviceName
        InputParam['DefaultGPUDeviceNumber']=np.uint32(DefaultGPUDeviceNumber)
        InputParam['ManualGroupSize']=ManualGroupSize
        InputParam['ManualLocalSize']=ManualLocalSize

        SolidFraction=None
        if SPP_ZONES>1:
            print('We will use SPP')
        else:
            InputParam['SPP_ZONES']=np.uint32(1)

        if SPP_VolumeFraction is not None:
            SolidFraction=np.zeros((N1+1,N2+1))
            SolidFraction[:N1,:N2]=SPP_VolumeFraction
            SolidFraction[0:NDelta,:]=0.
            SolidFraction[N1-NDelta:,:]=0.
            SolidFraction[:,0:NDelta]=0.
            SolidFraction[:,N2-NDelta:]=0.
            
            
        InputParam['SPP_ZONES']=np.uint32(SPP_ZONES)
        MultiZoneMaterialMap= PrepareSuperpositionArrays(MaterialMap2D,SolidFraction,SPP_ZONES=SPP_ZONES);
        InputParam['OrigMaterialMap']=MaterialMap2D
        InputParam['MaterialMap']=MultiZoneMaterialMap
        InputParam['SolidFraction']=SolidFraction


        print ('Matrix size= %i x %i  , spatial resolution = %g, time steps = %i, temporal step = %g, total sonication length %g ' %(N1,N2,h,TimeVector.size,dt,DurationSimulation))

        SensorOutput_orig,V,RMSValue,Snapshots_orig=self.ExecuteSimulation(InputParam,COMPUTING_BACKEND)# This will be executed either locally or remotely using Pyro4

        for n in range(len(SnapShots)):
            SnapShots[n]['V']=np.squeeze(Snapshots_orig[:,:,n])

        #SensorOutput_orig=np.sqrt(SensorOutput_orig); #Sensors captured the sum of squares of Vx, Vy and Vz

        RetValueSensors={}
        RetValueSensors['time']=TimeVector[SensorStart*SensorSubSampling:len(TimeVector):SensorSubSampling]
        for key,index in IndexSensorMaps.items():
            if index>=0:
                RetValueSensors[key]=SensorOutput_orig[:,0:len(RetValueSensors['time']),index]

        if (IntervalSnapshots>0):
            RetSnap=SnapShots
        else:
            RetSnap=[]

        #now time to organize this a dictionary
        RetValueRMS={}
        RetValuePeak={}

        if SelRMSorPeak==1 or SelRMSorPeak==3:
            for key,index in IndexRMSMaps.items():
                if index>=0:
                    #%in RMSValue we have the sum of square values over time, we need a
                    #%final calculation to have the real RMS
                    RetValueRMS[key]=np.sqrt(RMSValue[:,:,index,0]/(len(TimeVector)-SensorStart*SensorSubSampling))
        if SelRMSorPeak==2:
            for key,index in IndexRMSMaps.items():
                if index>=0:
                    RetValuePeak[key]=RMSValue[:,:,index,0]
        elif SelRMSorPeak==3:
            for key,index in IndexRMSMaps.items():
                if index>=0:
                    RetValuePeak[key]=RMSValue[:,:,index,1]

        pFactor=MaterialProperties[MaterialMap,0]*MaterialProperties[MaterialMap,1]**2/SpatialStep
        if 'Pressure' in RetValuePeak:
            #What was calculated in the low level function was the stencil gradient of each Vx, Vy, Vz 
            # We need to scale back to rho c^2 /Spatialstep
            RetValuePeak['Pressure']*=pFactor
            
        if 'Pressure' in RetValueRMS:
            RetValueRMS['Pressure']*= pFactor

        for pgrad in ['Pressure','Pressure_gx','Pressure_gy']:
            if pgrad in RetValueSensors:
                ii,jj=np.unravel_index(IndexSensors-1, SensorMap.shape, order='F')
                RetValueSensors[pgrad]*=np.repeat(pFactor[ii,jj].reshape([RetValueSensors[pgrad].shape[0],1]),
                                                    RetValueSensors[pgrad].shape[1],axis=1)
                if '_gx' in pgrad or '_gy' in pgrad:
                    RetValueSensors[pgrad]*=1.0/h

        if  IntervalSnapshots>0:
            if len(RetValueRMS)>0 and len(RetValuePeak)>0:
                return RetValueSensors,V,RetValueRMS,RetValuePeak,InputParam,RetSnap
            elif len(RetValueRMS)>0:
                return RetValueSensors,V,RetValueRMS,InputParam,RetSnap
            elif len(RetValuePeak)>0:
                return RetValueSensors,V,RetValuePeak,InputParam,RetSnap
            else:
                raise SystemError("How we got a condition where no RMS or Peak value was selected")
        else:
            if len(RetValueRMS)>0 and len(RetValuePeak)>0:
                return RetValueSensors,V,RetValueRMS,RetValuePeak,InputParam
            elif len(RetValueRMS)>0:
                return RetValueSensors,V,RetValueRMS,InputParam
            elif len(RetValuePeak)>0:
                return RetValueSensors,V,RetValuePeak,InputParam
            else:
                raise SystemError("How we got a condition where no RMS or Peak value was selected")



    def ExecuteSimulation(self,InputParam,COMPUTING_BACKEND):
        if COMPUTING_BACKEND == 1:
            print( "Performing Simulation with GPU CUDA")
            SensorOutput_orig,V,RMSValue,Snapshots_orig=StaggeredFDTD_2D_CUDA(InputParam)
        elif COMPUTING_BACKEND == 2:
            print ("Performing Simulation with GPU OPENCL")
            SensorOutput_orig,V,RMSValue,Snapshots_orig=StaggeredFDTD_2D_OPENCL(InputParam)
        elif COMPUTING_BACKEND == 3:
            print ("Performing Simulation with GPU METAL")
            SensorOutput_orig,V,RMSValue,Snapshots_orig=StaggeredFDTD_2D_METAL(InputParam)
        else:
            raise ValueError (" CPU not yet supported")
            
        return SensorOutput_orig,V,RMSValue,Snapshots_orig

    def CalculateMatricesForPropagation(self,MaterialMap, MaterialProperties, Frequency,QfactorCorrection,h,AlphaCFL,QCorrection=1.0):

        rho=MaterialProperties[:,np.int32(StaggeredConstants['ColDensity'])].flatten()
        VLong=MaterialProperties[:,np.int32(StaggeredConstants['ColLongSOS'])].flatten()
        VShear=MaterialProperties[:,np.int32(StaggeredConstants['ColShearSOS'])].flatten()
        ALong=MaterialProperties[:,np.int32(StaggeredConstants['ColLongAtt'])].flatten()
        AShear=MaterialProperties[:,np.int32(StaggeredConstants['ColShearAtt'])].flatten()

        UniqueMaterial=np.unique(MaterialMap.flatten())

        if np.max(UniqueMaterial) >= MaterialProperties.shape[0]:
            raise ValueError('The map in MaterialMap must have as many different values as materials identified in MaterialProperties (number of rows)');


        VShearUnique=VShear
        VLongUnique=VLong
        RhoUnique=rho
        ALongUnique=ALong
        AShearUnique=AShear

        dt,RhoMat,MiuMat, LambdaMiuMat, LambdaMat,TauLongMat,TauShearMat,TauSigmaMat,AnalysisQFactorLong,AnalysisQFactorShear=self.CalculateLambdaMiuMatrices(VLongUnique,VShearUnique,RhoUnique,ALongUnique,
                                                                                                                                AShearUnique,Frequency,QfactorCorrection,h,AlphaCFL,QCorrection=QCorrection)

        return dt,RhoMat,MiuMat, LambdaMiuMat, LambdaMat,TauLongMat,TauShearMat,TauSigmaMat,AnalysisQFactorLong,AnalysisQFactorShear

    def CalculateLambdaMiuMatrices(self,VLongInput,VShearInput,RhoMat,ALongInput,AShearInput,Frequency,QfactorCorrection,h,AlphaCFL,CheckOnlyParameters=False,QCorrection=1.0):
        Omega=Frequency*2*np.pi
        VMaxLong=np.max(VLongInput)


        #%% here comes the fun, calculate the relaxation coefficients,
        #%% first, we detected where attenuation is zero to avoid problems
        AttLongNonZero=ALongInput!=0.0
        AttShearNonZero=AShearInput!=0.0
        #%factor Qs, and Qp, is given by the number of wavelenghts required to
        #%attenuate the amplitude by exp(-pi), meaning alpha *Qdistance = pi;
        #% Blanch, Joakim O., Johan OA Robertsson, and William W. Symes. "Modeling of a constant Q: Methodology and algorithm for an efficient and optimally inexpensive viscoelastic technique." Geophysics 60.1 (1995): 176-184.

        #% We calculate Qs, Ql for a single relation mechanism, that is ok for
        #% single frequency relaxation :
        #%Bohlen, Thomas. "Parallel 3-D viscoelastic finite difference seismic modelling." Computers & Geosciences 28.8 (2002): 887-899.

        SubALong=ALongInput[AttLongNonZero].copy().flatten()
        SubAShear=AShearInput[AttShearNonZero].copy().flatten()

        QLong=np.pi/SubALong
        QShear=np.pi/SubAShear

        if QLong.shape[0]!=0:
            QLong=QLong/(VLongInput[AttLongNonZero]/Frequency)

        if QShear.shape[0]!=0:
            QShear=QShear/(VShearInput[AttShearNonZero]/Frequency)

        NoAttenuation=False
        if QShear.shape[0]==0 and QLong.shape[0]==0:
             NoAttenuation=True


        TauSigma=np.ones(ALongInput.size)/Omega


        #%We save the results of the curves of the Q factor, because if we need
        #%to analyze later that Q makes sense, this will be done postiori

        AnalysisQFactorLong,TauLong,TauSigma_l=CalculateRelaxationCoefficients(ALongInput,QLong,Frequency,QCorrection=QCorrection);
        AnalysisQFactorShear,TauShear,TauSigma_s=CalculateRelaxationCoefficients(AShearInput,QShear,Frequency,QCorrection=QCorrection);


        if QfactorCorrection: # %dispersion correction...
           Q_cw_factor_long= np.real(np.sqrt(1.0/(1.0 + (1j*Omega*TauSigma_l*TauLong)  /(1.0+1j*Omega*TauSigma_l))))
           Q_cw_factor_shear=np.real(np.sqrt(1.0/(1.0 + (1j*Omega*TauSigma_s*TauShear) /(1.0+1j*Omega*TauSigma_s))))


           print ("VLongInput,VShearInput", np.unique(VLongInput), np.unique(VShearInput))
           print ("Q_cw_factor_long,Q_cw_factor_shear", np.unique(Q_cw_factor_long), np.unique(Q_cw_factor_shear))
           VLongMat=VLongInput*Q_cw_factor_long
           VShearMat=VShearInput*Q_cw_factor_shear
           #", np.unique(VLongMat), np.unique(VShearMat))
        else:
           VLongMat=VLongInput.copy()
           VShearMat=VShearInput.copy()

        MiuMat = VShearMat**2*RhoMat
        LambdaMiuMat =  VLongMat**2*RhoMat
        LambdaMat = LambdaMiuMat - 2*MiuMat

        if CheckOnlyParameters:
            PoisonRatio=LambdaMat.flatten()/(2.0*(LambdaMat.flatten()+MiuMat.flatten()))
            return PoisonRatio


        # %%verify the time step condition,
        #%after
        #%Sun, Chengyu, Yunfei Xiao, Xingyao Yin, and Hongchao Peng. "Stability condition of finite difference solution for viscoelastic wave equations." Earthquake Science 22, no. 5 (2009): 479-485
        #%%I thinks there is a typo in the paper that stipulates
        #%% 4*h^4./(3*vp2_Long*SumAbsWeights) as part of the calculations of dt,
        #% but letting it like that would make no sense for cases where Q-->inf
        #% , in that situation, dt must converge as in other papers to *h*6/7/sqrt(3)/VMax
        #% after reviewing the paper equations, it is clear there is a typo,
        #% the previous equation (no mumber in the paper) to (17) is correct and
        #% and should translates to 4*h^2./(3*vp2_Long*SumAbsWeights)
        if NoAttenuation==False:
            WeighCoeff=np.array([ -1.7857143e-3, 2.5396825e-2,  -0.2,  1.6 ,-2.8472222 ,1.6,  -0.2,  2.5396825e-2, -1.7857143e-3])# %these are the coefficients for order 4-th
            SumAbsWeights=np.sum(np.abs(WeighCoeff))

            vp2_Long=(np.sqrt(QLong**2+1)+QLong)*VLongMat[AttLongNonZero]**2/(2.0*QLong)
            HLongCond=np.sqrt(Omega**2*h**4/(9*QLong**2.*vp2_Long**2 * SumAbsWeights**2) + 4*h**2/(3*vp2_Long*SumAbsWeights)) -\
                Omega*h**2/(3*QLong*vp2_Long*SumAbsWeights)

            vp2_Shear=(np.sqrt(QShear**2+1)+QShear)*VShearMat[AttShearNonZero]**2/(2*QShear)
            HShearCond=np.sqrt(Omega**2*h**4/(9*QShear**2*vp2_Shear**2 * SumAbsWeights**2) + 4*h**2/(3*vp2_Shear*SumAbsWeights)) -\
                Omega*h**2/(3*QShear*vp2_Shear*SumAbsWeights)


            #%% dt using the approach from Sun is slightly smaller, but it truly does the job, before, it was getting quickly unstable results,
            #%% making smaller dt manually helped, but now we have a better tuned approach
            dt=AlphaCFL*np.min(np.hstack((HLongCond.flatten(),HShearCond.flatten())))
            print (np.min(np.hstack((HLongCond.flatten(),HShearCond.flatten()))))

        else:
            dt=AlphaCFL*h*6.0/7.0/np.sqrt(3.0)/VMaxLong #%after: Bohlen, Thomas. "Parallel 3-D viscoelastic finite difference seismic modelling." Computers & Geosciences 28.8 (2002): 887-899.

        #print ("dt,VLongMat,VShearMat,TauLong,TauShear,TauSigma,VLongInput,VShearInput", dt, np.unique(VLongMat),  np.unique(VShearMat),np.unique(TauLong),np.unique(TauShear),np.unique(TauSigma),np.unique(VLongInput),np.unique(VShearInput))
        return dt,RhoMat,MiuMat, LambdaMiuMat, LambdaMat,TauLong,TauShear,TauSigma,AnalysisQFactorLong,AnalysisQFactorShear

    ###########################################################

########### AUXILIARY functions #######################

def EvalQ(x0,w,Tau_sigma):
    Tau=x0
    F=w*Tau_sigma*Tau/(1.0+w**2*Tau_sigma**2*(1.0+Tau))
    return F

def EvalQp(x0,w):
    Tau=x0[0]
    Tau_sigma=x0[1]
    F=w*Tau_sigma*Tau/(1.0+w**2*Tau_sigma**2*(1.0+Tau))
    return F

def I0l(TauSigma,w):
    F=1.0/2/TauSigma*np.log(1+w**2*TauSigma**2)
    return F


def I1l(TauSigma,w):
    F=1.0/2/TauSigma*(np.arctan(w*TauSigma)-w*TauSigma/(1+w**2*TauSigma**2))
    return F

def EvalTau(x0,w):
    LowFreq=w[0]
    HighFreq=w[1]
    TauSigma=x0;
    F=(I0l(TauSigma,HighFreq)-I0l(TauSigma,LowFreq))/(I1l(TauSigma,HighFreq)-I1l(TauSigma,LowFreq))
    return F

def OptimalTauForQFactor(QValue,CentralFreqHz):
#%%we implement at lin. sqr. root minimization of the Quality factor for the
#%%viscoelastic function with one relaxation mechanism
#%% Blanch, Joakim O., Johan OA Robertsson, and William W. Symes. "Modeling of a constant Q: Methodology and algorithm for an efficient and optimally inexpensive viscoelastic technique." Geophysics 60.1 (1995): 176-184.
#%% and
#%% Bohlen, Thomas. "Parallel 3-D viscoelastic finite difference seismic modelling." Computers & Geosciences 28.8 (2002): 887-899.

    LowFreq=CentralFreqHz-CentralFreqHz*0.05 #% we cover a bandwith of +/- 10% the central frequency
    HighFreq=CentralFreqHz+CentralFreqHz*0.05

    LowFreq=LowFreq*2*np.pi
    HighFreq=HighFreq*2*np.pi

    CentralFreq=CentralFreqHz*2*np.pi

    TauSigma=1.0/CentralFreq
    #%the formula is very good to give a initial guess
    #Tau=1.0/QValue*(I0l(TauSigma,HighFreq)-I0l(TauSigma,LowFreq))/(I1l(TauSigma,HighFreq)-I1l(TauSigma,LowFreq))
    Tau=2.0/QValue
    TauEpsilon = (Tau+1.0)*TauSigma
    #%x0=[TauSigma ,TauEpsilon ];
    x0=Tau
    SpectrumToValidate=np.linspace(LowFreq,HighFreq,num=100).flatten() #%fifty steps should be good

    QOptimal=1.0/QValue*np.ones((SpectrumToValidate.size,1))


    fh =(lambda x:np.sum((EvalQ(x,SpectrumToValidate,TauSigma)-QOptimal)**2))
    fhp =(lambda x:np.sum((EvalQp(x,SpectrumToValidate)-QOptimal)**2))


    x,fx,iuts,imode,smode = fmin_slsqp(fh,x0,bounds=[(0,np.inf)],full_output=True,iprint=0)

    xp,fxp,iutsp,imodep,smodep = fmin_slsqp(fhp,[Tau,TauSigma],bounds=[(0,1),(0,1)],full_output=True,iprint=0)

   # print('2.0/QValue, x0, x, xp, TauSigma',2.0/QValue,x0,x,xp,TauSigma)
    
    Tau=x[0]
    #Tau=xp[0]
    #Tau=2.0/QValue
    TauEpsilon = (Tau+1)*TauSigma
    Qres=1.0/EvalQ(Tau,SpectrumToValidate,TauSigma)

    Error_LSQ=np.sum((Qres-QValue)**2)/Qres.size

    #%%QValueFormula=QValue-QValue*0.1;% THIS IS TRULY AD HOC, as noted in Blanch, this has be to be done
    #%% to compensate effects of linearization, but yet,
    #%% lsqlin seems to do a good job, without having to sort out an kitchen formula... the formula is super sensitive to the range of frequencies  to be tested , which is not good at all

    fCal=[270e3,836e3,1402e3]
    Adj=[0.025,0.01,0.005]

    if CentralFreqHz<fCal[0] or CentralFreqHz>fCal[-1] :
        # warnings.warn('Central frequency (kHz) %f  outside the range of tested frequencies for adjustment of attenuation [%f,%f]' %\
        #                 (CentralFreqHz/1e3,fCal[0],fCal[-1]))
        if CentralFreqHz<fCal[0]:
            QValueFormula=QValue-QValue**Adj[0];
        else:
            QValueFormula=QValue-QValue**Adj[-1]
    else:
        intF = interp1d(fCal, Adj)
        QAdj=intF(CentralFreqHz)
        QValueFormula=QValue-QValue*QAdj
    


    TauSigmaFormula=1.0/CentralFreq;
    TauFormula=1.0/QValueFormula*(I0l(TauSigmaFormula,HighFreq)-I0l(TauSigmaFormula,LowFreq))/(I1l(TauSigmaFormula,HighFreq)-I1l(TauSigmaFormula,LowFreq));
    TauEpsilonFormula = (TauFormula+1)*TauSigmaFormula;
    QresFormula=1.0/EvalQ(TauFormula,SpectrumToValidate,TauSigmaFormula);
    Error_Formula=np.sum((QresFormula-QValue)**2)/Qres.size

    # print('TauFormula, TauSigmaFormula',TauFormula, TauSigmaFormula)

    return Tau,TauSigma,Qres,SpectrumToValidate,Error_LSQ
    
def CalculateRelaxationCoefficients(AttMat,Q,Frequency,QCorrection=1.0):

    AttNonZero=AttMat!=0
    if np.isscalar(QCorrection):
        Q*=QCorrection
    else:
        Q*=QCorrection[AttNonZero]

    IndAttNonZero=np.nonzero(AttNonZero.flatten().T)[0]
    AnalysisQFactor={}
    AnalysisQFactor['Spectrum']=None
    AnalysisQFactor['Attenuation']=AttMat[IndAttNonZero]
    AnalysisQFactor['Qres']=[]
    AnalysisQFactor['Qdesired']=Q
    AnalysisQFactor['Error_LSQ']=np.zeros((IndAttNonZero.size,1))
    TempTau=np.zeros(IndAttNonZero.size)
    TauSigma=np.zeros(IndAttNonZero.size)
    QresTemp=[]
    Error_LSQTemp=np.zeros(IndAttNonZero.size)
    SpectrumToValidate=None
    print('Calculating Relaxation coefficients for ',IndAttNonZero.size)
    print(' Q unique entries ',len(np.unique(Q)))
    
    for n in range(IndAttNonZero.size):
        Tau,Ts,Qres,SpectrumToValidateTemp,Error_LSQ=OptimalTauForQFactor(Q[n],Frequency)
        #%[Tau,Ts,Qres,SpectrumToValidateTemp,Error_LSQ]=OptimalTauAndTaueEpsForQFactor(Q(n),Frequency);
        if n==0:
            SpectrumToValidate=SpectrumToValidateTemp.copy()

        TempTau[n]=Tau
        TauSigma[n]=Ts
        QresTemp.append(Qres)
        Error_LSQTemp[n]=Error_LSQ

    Tau=np.zeros(AttMat.size)
    TauSigma_l=Tau.copy()
    Tau[IndAttNonZero]=TempTau
    TauSigma_l[IndAttNonZero]=TauSigma;
    AnalysisQFactor['Spectrum']=SpectrumToValidate;    #%the spectrum is the same for every value of Q
    AnalysisQFactor['Error_LSQ']=Error_LSQTemp;
    AnalysisQFactor['Qres']=QresTemp
    return AnalysisQFactor,Tau,TauSigma_l


def PrepareSuperpositionArrays(SourceMaterialMap,SolidFraction,SPP_ZONES=1,OrderExtra=2,bRemoveIsolatedRegions=True):
        #This function will assign the id of material in the fraction elements that are the closest to the water solid interfaces
        #if SPP_ZONES==1 is False, we just create dummy arrays, as these are need to be passed to the low level function for completeness
        #bRemoveIsolatedRegions controls if disconected regions will need to be removed as this can occur if SolidFraction was
        #calculated with no perfectly interfaces crossing the voxels (or small issues with whatever CSG method to caclulate intersection)

        ZoneCount=SPP_ZONES
        if ZoneCount>1 or SolidFraction is not None:
            SolidRegion=SourceMaterialMap!=0
            MaterialMap=SourceMaterialMap.copy()

            NewMaterialMap=SourceMaterialMap.copy()
            s=SourceMaterialMap.shape
            MultiZoneMaterialMap=np.zeros((s[0],s[1],ZoneCount),dtype=np.uint32)

            ExpandaMaterial=((SolidRegion)^(SolidFraction>0))
            ExpandaMaterial=((ExpandaMaterial)&(SolidFraction>0))
            ii,jj=np.where(ExpandaMaterial)

            mgrid = np.lib.index_tricks.nd_grid()
            iS,jS=mgrid[-1:2,-1:2]
            # print(ii.shape)
            for i,j in zip(ii,jj):
                ssi=i+iS
                ssj=j+jS
                
                SubMat=MaterialMap[ssi,ssj]
                assert(np.all(SubMat==0)==False)
                assert(np.any(SubMat==0))
                sel=SubMat!=0
                ssi=iS[sel]
                ssj=jS[sel]

                QuadDistance=np.linalg.norm(np.vstack((ssi,ssj)),axis=0)
                mIn=np.argmin(QuadDistance)
                assert(QuadDistance[mIn]!=0.0)
                SubMat=SubMat[sel]
                NewMaterialMap[i,j]=SubMat[mIn]

            SolidRegion=NewMaterialMap!=0

            SuperpositionMap = np.zeros(SourceMaterialMap.shape,dtype=np.uint8)
            SkullRingFraction=((SolidFraction>0)&(SolidFraction<=1.0))
            ExpandedRing=ndimage.binary_dilation(SkullRingFraction,iterations=SPP_ZONES)
            ExpandedRing[SolidFraction==1.0]=True
            SuperpositionMap[ExpandedRing]=1

            ExtraLayers=ndimage.binary_dilation(ExpandedRing,iterations=OrderExtra)
            ExtraLayers=np.logical_xor(ExtraLayers,ExpandedRing)
            SuperpositionMap[ExtraLayers]=2


            SelSuperpoistion=SuperpositionMap>0
            AllIndexesLarge=np.where(SelSuperpoistion)
            AllIndexesLargeFlat=np.where(SelSuperpoistion.flatten())[0]
            AllIndexesLarge=np.vstack((AllIndexesLarge[0].astype(np.uint32),AllIndexesLarge[1].astype(np.uint32),AllIndexesLarge[2].astype(np.uint32))).T

            MatMap_zone=  np.zeros((ZoneCount,AllIndexesLargeFlat.size),dtype=MaterialMap.dtype)
            SubMat=NewMaterialMap.flatten()[AllIndexesLargeFlat]
            SelFraction=SolidFraction.flatten()[AllIndexesLargeFlat]

            for zone in range(ZoneCount):
                frac=(zone)/ZoneCount
                MatMap_zone[zone,SelFraction>frac]=SubMat[SelFraction>frac]
                assert(np.sum(SelFraction>frac)==np.sum(MatMap_zone[zone,:]>0))
                ZoneMaterialMap=NewMaterialMap*0
                ZoneMaterialMap[SolidFraction>frac]=NewMaterialMap[SolidFraction>frac]

                MultiZoneMaterialMap[:,:,:,zone]=ZoneMaterialMap

                if bRemoveIsolatedRegions:
                    lab,num_features =ndimage.label(MultiZoneMaterialMap[:,:,:,zone])
                    if num_features>1: #only if we have more than 1 big block
                        LabSize=[]
                        for k in range(1,num_features+1):
                            LabSize.append([(lab==k).sum(),k])
                        LabSize.sort() #we sort by size, the largest skull region is at the end
                        LabSize=np.array(LabSize)
                        Reg=MultiZoneMaterialMap[:,:,:,zone]
                        Reg[lab!=LabSize[-1,1]]=0 #we made background all small regions
                        print('Removing %f %% of isolated voxels' % (LabSize[:-1,0].sum()/LabSize[-1,0]*100) )
                        MultiZoneMaterialMap[:,:,:,zone]=Reg
                    else:
                        print('No isolated voxels were found')

        else:
            s=SourceMaterialMap.shape
            MultiZoneMaterialMap=np.zeros((s[0],s[1],1),dtype=np.uint32)
            MultiZoneMaterialMap[:,:,0]=SourceMaterialMap
        return MultiZoneMaterialMap



from calendar import c
from multiprocessing.dummy import Array
from unicodedata import name
import numpy as np
import os
from pathlib import Path
import platform
import time
import sys
import gc
import tempfile
from shutil import copyfile

from .StaggeredFDTD_2D_With_Relaxation_BASE import StaggeredFDTD_2D_With_Relaxation_BASE

from distutils.sysconfig import get_python_inc
from math import ceil
import ctypes
from ctypes import c_byte, c_int, c_uint32, c_float, c_wchar_p, c_uint64

mc = None

def ListDevices():
    import metalcomputebabel 
    devicesIDs=[]
    devices = metalcomputebabel.get_devices()
    for dev in devices:
        devicesIDs.append(dev.deviceName)
    return devicesIDs

class StaggeredFDTD_2D_With_Relaxation_METAL_MetalCompute(StaggeredFDTD_2D_With_Relaxation_BASE):
    '''
    This version is mainly for Mx processors and which is based in a modified+forked version of metalcompute. As X64 will be phased out, in the future the metalcompute version will take over
    '''
    def __init__(self, arguments):
        global mc
        import metalcomputebabel
        mc = metalcomputebabel
        # Definition of some constants, etc
        self.MAX_SIZE_PML = 101
        self._c_mex_type = np.zeros(12, np.uint64)
        self._c_uint_type = np.uint64(0)
        self.HOST_INDEX_MEX = np.zeros((31, 2), np.uint64)
        self.HOST_INDEX_UINT = np.zeros((3, 2), np.uint64)
        self.LENGTH_INDEX_MEX = self.HOST_INDEX_MEX.shape[0]
        self.LENGTH_INDEX_UINT = self.HOST_INDEX_UINT.shape[0]
        self.ZoneCount = arguments['SPP_ZONES']
        self._IndexDataMetal = {
            "V_x_x":0, "V_y_x":0, "V_x_y":0, "V_y_y":0,
            "Vx":1, "Vy":1,
            "Rxx":2, "Ryy":2,
            "Rxy":3,
            "Sigma_x_xx":4, "Sigma_y_xx":4, "Sigma_x_yy":4, "Sigma_y_yy":4,
            "Sigma_x_xy":5, "Sigma_y_xy":5,
            "Sigma_xy":6, "Sigma_xx":6,
            "Sigma_yy":7, "Pressure":7,
            "SourceFunctions":8,
            "LambdaMiuMatOverH":9, "LambdaMatOverH":9, "MiuMatOverH":9,
            "TauLong":9, "OneOverTauSigma":9, "TauShear":9, "InvRhoMatH":9, "Ox":9, "Oy":9,  
            "SqrAcc":10,
            "SensorOutput":11
            }
        self.C_IND = {
            "IndexSensorMap":0, "SourceMap":1, "MaterialMap": 2,
            "nStep":0, "TypeSource":1, 
            "V_x_x":0, "V_y_x":1, "V_x_y":2, "V_y_y":3, "Vx":4, "Vy":5,
            "Rxx":6, "Ryy":7, "Rxy":8,
            "Sigma_x_xx":9, "Sigma_y_xx":10, "Sigma_x_yy":11, "Sigma_y_yy":12,
            "Sigma_x_xy":13, "Sigma_y_xy":14, "Sigma_xy":15, 
            "Sigma_xx":16, "Sigma_yy":17, 
            "SourceFunctions":18, "LambdaMiuMatOverH":19, 
            "LambdaMatOverH":20, "MiuMatOverH":21, 
            "TauLong":22, "OneOverTauSigma":23, "TauShear":24, 
            "InvRhoMatH":25, "Ox":26, "Oy":27,
            "Pressure":28, "SqrAcc":29, "SensorOutput":30, 
            }

        self.FUNCTION_LOCALS = {}
        for i in ['MAIN_1' ]:
            self.FUNCTION_LOCALS[i] = {'STRESS':[0, 0, 0], 'PARTICLE':[0, 0, 0]}
        self.FUNCTION_GLOBALS = {}
        for i in ['MAIN_1']:
            self.FUNCTION_GLOBALS[i] = {'STRESS':[0, 0, 0], 'PARTICLE':[0, 0, 0]}
                
        self.LENGTH_CONST_UINT = 2
        # self.LENGTH_CONST_MEX = 1+self.MAX_SIZE_PML*4

        extra_params = {"BACKEND":"METAL"}
        super().__init__(arguments, extra_params)
        
    def _PostInitScript(self, arguments, extra_params):
        print("Attempting Metal Initiation...")
        devices = mc.get_devices()
        SelDevice=None
        for n,dev in enumerate(devices):
            if arguments['DefaultGPUDeviceName'] in dev.deviceName:
                SelDevice=dev
                break
        if SelDevice is None:
            raise SystemError("No Metal device containing name [%s]" %(arguments['DefaultGPUDeviceName']))
        else:
            print('Selecting device: ', dev.deviceName)
        SCode = []
        SCode.append("#define mexType " + extra_params['td'] +"\n")
        SCode.append("#define METAL\n")
        SCode.append("#define METALCOMPUTE\n")
        SCode.append("#define MAX_SIZE_PML 101\n")
        extra_params['SCode'] = SCode
        self.ctx = mc.Device(n)
        self.ConstantBufferUINT=np.zeros(self.LENGTH_CONST_UINT,np.uint32)
        # self.ConstantBufferMEX=np.zeros(self.LENGTH_CONST_MEX,np.float32)
        print(self.ctx)
        if 'arm64' not in platform.platform():
            print('Setting Metal for External or AMD processor')
            self.ctx.set_external_gpu(1) 
    
    def _InitSymbol(self, IP,_NameVar,td,SCode):
        if td in ['float','double']:
            res = 'constant ' + td  + ' ' + _NameVar + ' = %0.9g;\n' %(IP[_NameVar])
        else:
            lType =' _PT '
            res = 'constant '+ lType  + _NameVar + ' = %i;\n' %(IP[_NameVar])
        SCode.append(res)
        
    def _InitSymbolArray(self, IP,_NameVar,td,SCode):
        res =  "constant "+ td + " " + _NameVar + "_pr[%i] ={\n" % (IP[_NameVar].size)
        for n in range(IP[_NameVar].size):
            if td in ['float','double']:
                res+="%.9g" % (IP[_NameVar][n])
            else:
                res+="%i" % (IP[_NameVar][n])
            if n<IP[_NameVar].size-1:
                res+=',\n'
            else:
                res+='};\n'
        SCode.append(res)   


    def _UpdateSymbol(self, IP, _NameVar,td, SCode):
        if td == "float":
            self.constant_buffer_mex.modify(np.array([IP[_NameVar]],dtype=np.float32),int(self.C_IND[_NameVar]),1)
        elif td == "unsigned int": 
            self.constant_buffer_uint.modify(np.array(IP[_NameVar],dtype=np.uint32),int(self.C_IND[_NameVar]),1)
        else:
            raise ValueError("Something was passed incorrectly in symbol initiation.")


    def _ownGpuCalloc(self, Name,td,dims,ArraysGPUOp):
        print("Allocating for", Name, dims, "elements")
        if Name == "Snapshots":
            pass
        elif td == "float":
            self.HOST_INDEX_MEX[self.C_IND[Name]][0] = np.uint64(self._c_mex_type[self._IndexDataMetal[Name]])
            self.HOST_INDEX_MEX[self.C_IND[Name]][1] = np.uint64(dims ) 
            self._c_mex_type[self._IndexDataMetal[Name]] += dims
        elif td == "unsigned int":
            self.HOST_INDEX_UINT[self.C_IND[Name]][0] = np.uint64(self._c_uint_type)
            self.HOST_INDEX_UINT[self.C_IND[Name]][1] = np.uint64(dims)
            self._c_uint_type += dims

    
    def _CreateAndCopyFromMXVarOnGPU(self, Name,ArraysGPUOp,ArrayResCPU,flags=[]):
        print("Allocating for", Name, ArrayResCPU[Name].size, "elements")
        SizeCopy = ArrayResCPU[Name].size
        if Name in ['LambdaMiuMatOverH','LambdaMatOverH','MiuMatOverH','TauLong','OneOverTauSigma','TauShear','InvRhoMatH', 'Ox','Oy', 'SourceFunctions', 'SensorOutput','SqrAcc']: # float
            self.HOST_INDEX_MEX[self.C_IND[Name]][0] = np.uint64(self._c_mex_type[self._IndexDataMetal[Name]])
            self.HOST_INDEX_MEX[self.C_IND[Name]][1] = np.uint64(SizeCopy)
            self._c_mex_type[self._IndexDataMetal[Name]] += SizeCopy
        elif Name in ['IndexSensorMap','SourceMap','MaterialMap',]: # unsigned int
            self.HOST_INDEX_UINT[self.C_IND[Name]][0] = np.uint64(self._c_uint_type)
            self.HOST_INDEX_UINT[self.C_IND[Name]][1] = np.uint64(SizeCopy)
            self._c_uint_type += SizeCopy
    
    def _PreExecuteScript(self, arguments, ArraysGPUOp, outparams):
        print("Float entries:", np.sum(self._c_mex_type), "int entries:", self._c_uint_type)
        self.mex_buffer=[]
        for nSizes in self._c_mex_type:
            handle=self.ctx.buffer(nSizes*4)
            self.mex_buffer.append(handle)
        self.uint_buffer=self.ctx.buffer(self._c_uint_type*4)
        self.constant_buffer_uint=self.ctx.buffer(self.ConstantBufferUINT)

        self._IndexManip()

        for k in ['LambdaMiuMatOverH','LambdaMatOverH','MiuMatOverH','TauLong','OneOverTauSigma','TauShear','InvRhoMatH',\
                    'Ox','Oy','SourceFunctions']:
            self._CompleteCopyToGPU(k, arguments, arguments[k].size, "float")

        for k in ['IndexSensorMap','SourceMap','MaterialMap']:
            self._CompleteCopyToGPU(k, arguments, arguments[k].size, "unsigned int")

        if arguments['ManualLocalSize'][0]!=-1:
            self._SET_USER_LOCAL(arguments['ManualLocalSize'])
        else:
            self._CALC_USER_LOCAL("MAIN_1", "STRESS")
            self._CALC_USER_LOCAL("MAIN_1", "PARTICLE")
               
        if arguments['ManualGroupSize'][0] != -1:
            self._SET_USER_GLOBAL(arguments['ManualGroupSize'])
        else:
            self._CALC_USER_GROUP_MAIN(arguments, outparams)

        self.localSensor = [1, 1, 1]
        self.globalSensor = [ceil(arguments['IndexSensorMap'].size / self.localSensor[0]), 1, 1]

    def _IndexManip(self):
        index=np.zeros((self.LENGTH_INDEX_MEX,2),np.uint32)
        for i in range(self.LENGTH_INDEX_MEX):
            index[i,0] = np.uint32(np.uint64(0xFFFFFFFF) & np.uint64(self.HOST_INDEX_MEX[i][0])) # Not exactly sure if this works
            index[i,1] = np.uint32(np.uint64([self.HOST_INDEX_MEX[i][0]])>>32)
        self.index_mex=self.ctx.buffer(index)
        
        index=np.zeros((self.LENGTH_INDEX_UINT,2),np.uint32)
       
        for i in range(self.LENGTH_INDEX_UINT):
            index[i,0] = np.uint32(0xFFFFFFFF) & np.uint32(self.HOST_INDEX_UINT[i][0]) # Not exactly sure if this works
            index[i,1] = np.uint32(np.uint64([self.HOST_INDEX_MEX[i][0]])>>32)
        self.index_uint=self.ctx.buffer(index)


    def _CompleteCopyToGPU(self, Name, args, SizeCopy, td):
        if td == "float":
            self.mex_buffer[self._IndexDataMetal[Name]].modify(args[Name].flatten(order='F'),int(self.HOST_INDEX_MEX[self.C_IND[Name]][0]),int(SizeCopy))
        elif td == "unsigned int":
            self.uint_buffer.modify(args[Name].flatten(order='F'),int(self.HOST_INDEX_UINT[self.C_IND[Name]][0]),int(SizeCopy))
        else:
            raise RuntimeError("Something has gone horribly wrong.")
    

    def _CALC_USER_LOCAL(self, Name, Type):
        #this will be calculated by metalcompute library
        self.FUNCTION_LOCALS[Name][Type][0] = 1
        self.FUNCTION_LOCALS[Name][Type][1] = 1
        self.FUNCTION_LOCALS[Name][Type][2] = 1
        # print(Name, "local", Type + " = [" + str(w) + ", " + str(h) + ", " + str(z) + "]")
    
    def _SET_USER_LOCAL(self, ManualLocalSize):
        for Type in ['STRESS', 'PARTICLE']:
            for index in range(2):
                self.FUNCTION_LOCALS['MAIN_1'][Type][index] = ManualLocalSize[index] # Can probably change this
    
    def _SET_USER_GLOBAL(self, ManualGlobalSize):
        for Type in ['STRESS', 'PARTICLE']:
            for index in range(2):
               self.FUNCTION_GLOBALS['MAIN_1'][Type][index] = ManualGlobalSize[index] 

    def _CALC_USER_GROUP_MAIN(self, arguments, outparams):
        self._outparams = outparams
        for Type in ['STRESS', 'PARTICLE']:
            for index in range(2):
                self.FUNCTION_GLOBALS['MAIN_1'][Type][index] = ceil((arguments[('N'+str(index + 1))]) / self.FUNCTION_LOCALS['MAIN_1'][Type][index])
            print("MAIN_1_global_" + Type, "=", str(self.FUNCTION_GLOBALS['MAIN_1'][Type]))
    

    def _InitiateCommands(self, AllC):
        prg = self.ctx.kernel(AllC)
        PartsStress=['MAIN_1']
        self.AllStressKernels={}
        for k in PartsStress:
            self.AllStressKernels[k]=prg.function(k+"_StressKernel")

        PartsParticle=['MAIN_1']
        self.AllParticleKernels={}
        for k in PartsParticle:
            self.AllParticleKernels[k]=prg.function(k+"_ParticleKernel")
        
        self.SensorsKernel=prg.function('SensorsKernel')

    def _Execution(self, arguments, ArrayResCPU, ArrayResOP):
        TimeSteps = arguments['TimeSteps']
        InitDict = {'nStep':0, 'TypeSource':int(arguments['TypeSource'])}
        outparams=self._outparams
        DimsKernel={}
        DimsKernel['MAIN_1']=[outparams['N1'],
                            outparams['N2']]
    
        nref=sys.getrefcount(self.mex_buffer[7])
        print('before exec nref',nref)
        for nStep in range(TimeSteps):
            InitDict["nStep"] = nStep
            for i in ['nStep', 'TypeSource']:
                self._UpdateSymbol(InitDict, i, 'unsigned int', [])

            self.ctx.init_command_buffer()
            AllHandles=[]
  
            for i in ["MAIN_1"]:
                nSize=np.prod(DimsKernel[i])
                handle=self.AllStressKernels[i](nSize,self.constant_buffer_uint,
                                               self.index_mex,
                                               self.index_uint, 
                                               self.uint_buffer,
                                               self.mex_buffer[0],
                                               self.mex_buffer[1],
                                               self.mex_buffer[2],
                                               self.mex_buffer[3],
                                               self.mex_buffer[4],
                                               self.mex_buffer[5],
                                               self.mex_buffer[6],
                                               self.mex_buffer[7],
                                               self.mex_buffer[8],
                                               self.mex_buffer[9],
                                               self.mex_buffer[10],
                                               self.mex_buffer[11])
                AllHandles.append(handle)

            for i in ["MAIN_1"]:
                nSize=np.prod(DimsKernel[i])
                handle = self.AllParticleKernels[i](nSize,self.constant_buffer_uint,
                                               self.index_mex,
                                               self.index_uint, 
                                               self.uint_buffer,
                                               self.mex_buffer[0],
                                               self.mex_buffer[1],
                                               self.mex_buffer[2],
                                               self.mex_buffer[3],
                                               self.mex_buffer[4],
                                               self.mex_buffer[5],
                                               self.mex_buffer[6],
                                               self.mex_buffer[7],
                                               self.mex_buffer[8],
                                               self.mex_buffer[9],
                                               self.mex_buffer[10],
                                               self.mex_buffer[11])

                AllHandles.append(handle)
            if (nStep % arguments['SensorSubSampling'])==0  and (int(nStep/arguments['SensorSubSampling'])>=arguments['SensorStart']):
                handle=self.SensorsKernel(np.prod(self.globalSensor),
                                self.constant_buffer_uint,
                                self.index_mex,
                                self.index_uint, 
                                self.uint_buffer,
                                self.mex_buffer[0],
                                self.mex_buffer[1],
                                self.mex_buffer[2],
                                self.mex_buffer[3],
                                self.mex_buffer[4],
                                self.mex_buffer[5],
                                self.mex_buffer[6],
                                self.mex_buffer[7],
                                self.mex_buffer[8],
                                self.mex_buffer[9],
                                self.mex_buffer[10],
                                self.mex_buffer[11])
                AllHandles.append(handle)
            self.ctx.commit_command_buffer()
            self.ctx.wait_command_buffer()
            while len(AllHandles)>0:
                handle = AllHandles.pop(0) 
                del handle
        # nref=sys.getrefcount(self.mex_buffer[7])
        # print('after exec nref',nref)
        if 'arm64' not in platform.platform():
            self.ctx.sync_buffers((self.mex_buffer[0],
                                    self.mex_buffer[1],
                                    self.mex_buffer[2],
                                    self.mex_buffer[3],
                                    self.mex_buffer[4],
                                    self.mex_buffer[5],
                                    self.mex_buffer[6],
                                    self.mex_buffer[7],
                                    self.mex_buffer[8],
                                    self.mex_buffer[9],
                                    self.mex_buffer[10],
                                    self.mex_buffer[11]))
        for i in ['SqrAcc', 'SensorOutput']:
            SizeCopy = ArrayResCPU[i].size
            Shape = ArrayResCPU[i].shape
            print('getting ',i,self._IndexDataMetal[i])
            Buffer=np.frombuffer(self.mex_buffer[self._IndexDataMetal[i]],dtype=np.float32)[int(self.HOST_INDEX_MEX[self.C_IND[i]][0]):int(self.HOST_INDEX_MEX[self.C_IND[i]][0]+SizeCopy)]
            ArrayResCPU[i][:,:,:] = Buffer.reshape(Shape,order='F').copy()
     
        
        SizeBuffer = {1:0, 6:0, 7:0, 9:0}
        for i in ['Vx', 'Vy',  'Sigma_xx', 'Sigma_yy', 'Sigma_xy']:
            SizeBuffer[self._IndexDataMetal[i]] += ArrayResCPU[i].size* self.ZoneCount
        
        for i in ["LambdaMiuMatOverH", "LambdaMatOverH", "MiuMatOverH", "TauLong", "OneOverTauSigma", "TauShear", "InvRhoMatH", "Ox", "Oy"]:
            SizeBuffer[9] += arguments[i].size
        
        SizeBuffer[9] += ArrayResCPU['Pressure'].size* self.ZoneCount

        for i in ['Vx', 'Vy',  'Sigma_xx', 'Sigma_yy', 'Sigma_xy']:
            SizeCopy = ArrayResCPU[i].size * self.ZoneCount
            sz=ArrayResCPU[i].shape
            Shape = (sz[0],sz[1],self.ZoneCount)
            Buffer=np.frombuffer(self.mex_buffer[self._IndexDataMetal[i]],dtype=np.float32)[int(self.HOST_INDEX_MEX[self.C_IND[i]][0]):int(self.HOST_INDEX_MEX[self.C_IND[i]][0]+SizeCopy)]
            Buffer=Buffer.reshape(Shape,order='F')
            ArrayResCPU[i][:,:] = np.sum(Buffer,axis=2)/self.ZoneCount
        nref=sys.getrefcount(self.mex_buffer[7])
        # print('after 1st copying',nref)
        # print(gc.get_referrers(self.mex_buffer[7]))
        for i in ['Pressure']:
            SizeCopy = ArrayResCPU[i].size * self.ZoneCount
            sz=ArrayResCPU[i].shape
            Shape = (sz[0],sz[1],self.ZoneCount)
            Buffer=np.frombuffer(self.mex_buffer[self._IndexDataMetal[i]],dtype=np.float32)[int(self.HOST_INDEX_MEX[self.C_IND[i]][0]):int(self.HOST_INDEX_MEX[self.C_IND[i]][0]+SizeCopy)]
            Buffer=Buffer.reshape(Shape,order='F')
            ArrayResCPU[i][:,:] = np.sum(Buffer,axis=2)/self.ZoneCount
        # nref=sys.getrefcount(self.mex_buffer[7])
        # print('after 2nd copying',nref)
        # print(gc.get_referrers(self.mex_buffer[7]))
        # print(gc.get_referrers(self.mex_buffer[8]))
        del self.constant_buffer_uint
        # del self.constant_buffer_mex
        del self.index_mex
        del self.index_uint 
        del self.uint_buffer
        while len(self.mex_buffer)>0:
            handle = self.mex_buffer.pop(0)
            # nref=sys.getrefcount(handle)
            # print('mex nref',nref)
            del handle

def StaggeredFDTD_2D_METAL(arguments):
    Instance = StaggeredFDTD_2D_With_Relaxation_METAL_MetalCompute(arguments)
    Results = Instance.Results
    return Results



import numpy as np
import os
os.environ['GPU_FORCE_64BIT_PTR'] ="1"
from pathlib import Path

import pyopencl as cl
import time
from shutil import copyfile
import tempfile

from .StaggeredFDTD_3D_With_Relaxation_BASE import StaggeredFDTD_3D_With_Relaxation_BASE

#we will generate the _kernel-opencl.c file when importing
from distutils.sysconfig import get_python_inc
import platform

TotalAllocs=0
AllC=''

def ListDevices():
    devicesIDs=[]
    platforms = cl.get_platforms()
    for p in platforms:
        devices=p.get_devices() 
        for device in devices:
            devicesIDs.append(device.name)
    return devicesIDs

class StaggeredFDTD_3D_With_Relaxation_OPENCL(StaggeredFDTD_3D_With_Relaxation_BASE):
    def __init__(self, arguments):
        extra_params = {"BACKEND":"OPENCL"}
        super().__init__(arguments, extra_params)

    def _PostInitScript(self, arguments, extra_params):
        global TotalAllocs
        TotalAllocs=0
        SCode = []
        platforms = cl.get_platforms()
        bFound=False
        for p in platforms:
            devices=p.get_devices()  
            for device in devices:
                if arguments['DefaultGPUDeviceName'] in device.name:
                    bFound=True
                    break
            if bFound: break
        
        if bFound:
            print('Device ', arguments['DefaultGPUDeviceName'], ' Found!')
        else:
            raise ValueError('Device ' + arguments['DefaultGPUDeviceName'] + ' not found!')
        
        address_bits=device.get_info(cl.device_info.ADDRESS_BITS)

        print('Address bits', address_bits)

        if address_bits==32:
            SCode.append("#define _PT_32\n")
        SCode.append("#define mexType " + extra_params['td'] +"\n")
        SCode.append("#define OPENCL\n")
        extra_params['SCode'] = SCode

        self.ctx=cl.Context([device])
        self.queue = cl.CommandQueue(self.ctx)


    def _InitSymbol(self, IP,_NameVar,td,SCode):
        if td in ['float','double']:
            res = '__constant ' + td  + ' ' + _NameVar + ' = %0.9g;\n' %(IP[_NameVar])
        else:
            lType =' _PT '
            res = '__constant '+ lType  + _NameVar + ' = %i;\n' %(IP[_NameVar])
        SCode.append(res)
        
    def _InitSymbolArray(self, IP,_NameVar,td,SCode):
        res =  "__constant "+ td + " gpu" + _NameVar + "pr[%i] ={\n" % (IP[_NameVar].size)
        for n in range(IP[_NameVar].size):
            if td in ['float','double']:
                res+="%.9g" % (IP[_NameVar][n])
            else:
                res+="%i" % (IP[_NameVar][n])
            if n<IP[_NameVar].size-1:
                res+=',\n'
            else:
                res+='};\n'
        SCode.append(res)            
        
    def _ownGpuCalloc(self, Name,td,dims,ArraysGPUOp,flags=cl.mem_flags.READ_WRITE):
        global TotalAllocs
        if td in ['float','unsigned int']:
            f=4
        else: # double
            f=8
        print('Allocating for',Name,dims,'elements')
        ArraysGPUOp[Name]=cl.Buffer(self.ctx, flags,size=dims*f)
        TotalAllocs+=1            

    def _CreateAndCopyFromMXVarOnGPU(self, Name,ArraysGPUOp,ArrayResCPU, flags=cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR):
        global TotalAllocs
        print('Allocating for',Name,ArrayResCPU[Name].size,'elements')
        ArraysGPUOp[Name]=cl.Buffer(self.ctx, flags,hostbuf=ArrayResCPU[Name])
        TotalAllocs+=1

    def _Execution(self, arguments, ArrayResCPU, ArraysGPUOp):
        TimeSteps=arguments['TimeSteps']
        NumberSensors=arguments['IndexSensorMap'].size

        for nStep in range(TimeSteps):
            for k in self.AllStressKernels:
                self.AllStressKernels[k].set_arg(54,np.uint32(nStep))
                self.AllStressKernels[k].set_arg(55,arguments['TypeSource'])
            for k in self.AllParticleKernels:
                self.AllParticleKernels[k].set_arg(54,np.uint32(nStep))
                self.AllParticleKernels[k].set_arg(55,arguments['TypeSource'])
            for k in self.AllStressKernels:
                ev = cl.enqueue_nd_range_kernel(self.queue, self.AllStressKernels[k], self.AllGroupSizes[k], self.LocalSize)
            # queue.finish()
            for k in self.AllParticleKernels:
                ev = cl.enqueue_nd_range_kernel(self.queue, self.AllParticleKernels[k], self.AllGroupSizes[k], self.LocalSize)
            # queue.finish()
            if (nStep % arguments['SensorSubSampling'])==0  and (int(nStep/arguments['SensorSubSampling'])>=arguments['SensorStart']):
                self.SensorsKernel.set_arg(56,np.uint32(nStep))
                ev = cl.enqueue_nd_range_kernel(self.queue, self.SensorsKernel, (NumberSensors,1), None)
            self.queue.finish()


        bFirstCopy=True
        events=[]
        for k in ['Vx','Vy','Vz','Sigma_xx','Sigma_yy','Sigma_zz',
            'Sigma_xy','Sigma_xz','Sigma_yz','Pressure']:
            sz=ArrayResCPU[k].shape
            tempArray=np.zeros((sz[0],sz[1],sz[2],arguments['SPP_ZONES']),dtype=ArrayResCPU[k].dtype,order='F')
            events.append(cl.enqueue_copy(self.queue, tempArray, ArraysGPUOp[k]))
            cl.wait_for_events(events)
            ArrayResCPU[k][:,:,:]=tempArray.sum(axis=3)/arguments['SPP_ZONES']
            
        for k in ['SqrAcc','Snapshots','SensorOutput']:
            events.append(cl.enqueue_copy(self.queue, ArrayResCPU[k], ArraysGPUOp[k]))
        cl.wait_for_events(events)
          
        self.queue.finish()
        
        
        for k in ArraysGPUOp:
            ArraysGPUOp[k].release()


    def _PreExecuteScript(self, arguments, ArraysGPUOp, dummy):
        self.TimeSteps = arguments['TimeSteps']
        N1=arguments['N1']
        N2=arguments['N2']
        N3=arguments['N3']
        _IndexDataKernel=["V_x_x",
            "V_y_x",
            "V_z_x",
            "V_x_y",
            "V_y_y",
            "V_z_y",
            "V_x_z",
            "V_y_z",
            "V_z_z",
            "Vx",
            "Vy",
            "Vz",
            "Rxx",
            "Ryy",
            "Rzz",
            "Rxy",
            "Rxz",
            "Ryz",
            "Sigma_x_xx",
            "Sigma_y_xx",
            "Sigma_z_xx",
            "Sigma_x_yy",
            "Sigma_y_yy",
            "Sigma_z_yy",
            "Sigma_x_zz",
            "Sigma_y_zz",
            "Sigma_z_zz",
            "Sigma_x_xy",
            "Sigma_y_xy",
            "Sigma_x_xz",
            "Sigma_z_xz",
            "Sigma_y_yz",
            "Sigma_z_yz",
            "Sigma_xy",
            "Sigma_xz",
            "Sigma_yz",
            "Sigma_xx",
            "Sigma_yy",
            "Sigma_zz",
            "SourceFunctions",
            "LambdaMiuMatOverH",
            "LambdaMatOverH",
            "MiuMatOverH",
            "TauLong",
            "OneOverTauSigma",
            "TauShear",
            "InvRhoMatH",
            "SqrAcc",
            "MaterialMap",
            "SourceMap",
            "Ox",
            "Oy",
            "Oz",
            "Pressure"]
        assert len(_IndexDataKernel)==54
        
        for n,k in enumerate(_IndexDataKernel):
            for k2 in self.AllStressKernels:
                self.AllStressKernels[k2].set_arg(n,ArraysGPUOp[k])
            for k2 in self.AllParticleKernels:
                self.AllParticleKernels[k2].set_arg(n,ArraysGPUOp[k])

            self.SensorsKernel.set_arg(n,ArraysGPUOp[k])
        self.SensorsKernel.set_arg(54,ArraysGPUOp['SensorOutput'])
        self.SensorsKernel.set_arg(55,ArraysGPUOp['IndexSensorMap'])

        if arguments['ManualGroupSize'][0]!=-1:
            GroupSize=(arguments['ManualGroupSize'][0],arguments['ManualGroupSize'][1],arguments['ManualGroupSize'][2])
        else:
            GroupSize=(N1,N2,N3)

        self.AllGroupSizes={}
        self.AllGroupSizes['MAIN_1']=GroupSize
        

        if arguments['ManualLocalSize'][0]!=-1:
            self.LocalSize=(arguments['ManualLocalSize'][0],arguments['ManualLocalSize'][1],arguments['ManualLocalSize'][2])
        else:
            self.LocalSize=None


    def _InitiateCommands(self, SCode):
        AllC=''.join(SCode)
        prg = cl.Program(self.ctx,AllC).build()
        PartsStress=['MAIN_1']
        self.AllStressKernels={}
        for k in PartsStress:
            self.AllStressKernels[k]=getattr(prg,k+"_StressKernel")

        PartsParticle=['MAIN_1']
        self.AllParticleKernels={}
        for k in PartsParticle:
            self.AllParticleKernels[k]=getattr(prg,k+"_ParticleKernel")
        
        self.SensorsKernel=prg.SensorsKernel
        
    def CreateResults(self, ArrayResCPU):
        self.Results = ArrayResCPU['SensorOutput'],\
                {'Vx':ArrayResCPU['Vx'],\
                'Vy':ArrayResCPU['Vy'],\
                'Vz':ArrayResCPU['Vz'],\
                'Sigma_xx':ArrayResCPU['Sigma_xx'],\
                'Sigma_yy':ArrayResCPU['Sigma_yy'],\
                'Sigma_zz':ArrayResCPU['Sigma_zz'],\
                'Sigma_xy':ArrayResCPU['Sigma_xy'],\
                'Sigma_yz':ArrayResCPU['Sigma_yz'],\
                'Pressure':ArrayResCPU['Pressure']},\
                ArrayResCPU['SqrAcc'],ArrayResCPU['Snapshots']  



def StaggeredFDTD_3D_OPENCL(arguments):
    Instance = StaggeredFDTD_3D_With_Relaxation_OPENCL(arguments)
    return Instance.Results



import os
from glob import glob

import sys
if sys.version_info > (3, 2):
    unicode=str
    def bCheckIfStr(v):
        return type(v) is str
    def cStr(v):
        if type(v) is bytes:
            return v.decode("utf-8")
        else:
            return v
else:
    def bCheckIfStr(v):
        return (type(v) is str or type(v) is unicode)
    def cStr(v):
        return v
    
def bCheckdNumpyStr(v):
    if numpy.__version__ < "2.0.0":
        return type(v) is numpy.string_ or type(v) is numpy.unicode_ 
    else:
        return type(v) is numpy.str_

'''
Generic BackEnd to save hdf5 files using h5py, it seems more efficient than Pytables for "very-general-purpose saving". For better designed groups and datasets, pytables is way better

'''
import numpy
import h5py
import hdf5plugin 
from pydicom.uid import UID
from collections import OrderedDict

def ProcType(k,v,f,compatibility,complevel,group):
    nametype=type(v).__name__
    if group is None:
        group=f['/']
    if  type(v) in [dict,OrderedDict]:
        if isinstance(k,str):
            indexType='str'
        elif isinstance(k,unicode):
            if sys.version_info <= (3, 2):
                indexType='unicode'
            else:
                indexType='str'
        elif isinstance(k,int) or isinstance(k,numpy.int32) or isinstance(k,numpy.int16) or isinstance(k,numpy.int64):
            indexType='int'
        elif isinstance(k,float) or isinstance(k,numpy.float32) or isinstance(k,numpy.float64):
            indexType='float'
        else:
            raise TypeError('The type of key ' +str(k)+ ' is not supported ( ' + str(type(k)) + ' )')
        stk = k if (isinstance(k,str) or isinstance(k,unicode)) else str(k)
        newgroup=group.create_group(stk)
        newgroup.attrs["type"]="dict" if type(v) is dict else "OrderedDict"
        newgroup.attrs["type_key"]=indexType
        SaveToH5py(v,f,compatibility,complevel,newgroup)
    elif type(v) is list:
        newgroup=group.create_group(k)
        newgroup.attrs["type"]="list";
        for n in range(len(v)):
            vlist=v[n]
            itemname = "item_%d"%n
            ProcType(itemname,vlist,f,compatibility,complevel,newgroup)
    elif type(v)  is tuple:
        nameList = k
        newgroup=group.create_group(k)
        newgroup.attrs["type"]="tuple"
        for n in range(len(v)):
            vlist=v[n]
            itemname = "item_%d"%n
            ProcType(itemname,vlist,f,compatibility,complevel,newgroup)
    elif type(v)  is  numpy.ndarray :
        ##we can apply compression rules for larger arrays
        if v.nbytes >2**10: # if array takes more 1024 bytes
            if compatibility=="lzf":
                ds=group.create_dataset(k, data=v,compression="lzf",chunks=True,shuffle=True) #this is very simple but accessible compressor
            elif compatibility=="gzip":
                ds=group.create_dataset(k, data=v,compression="gzip",compression_opts=complevel,chunks=True,shuffle=True) #this gives a good compression and it is Matlab compatible
            elif compatibility=="blosc":
                ds=group.create_dataset(k, data=v, **hdf5plugin.Blosc(clevel=complevel),chunks=True) #this gives a good compression and it is blasting fast
        else:
            ds=group.create_dataset(k, data=v)
        ds.attrs["type"]="ndarray"
    elif  bCheckIfStr(v) or bCheckdNumpyStr(v):
        ds=group.create_dataset(k, data=v)
        ds.attrs["type"]=str(type(v))
    elif numpy.isscalar(v):
        ds=group.create_dataset(k, data=v)
        ds.attrs["type"]="scalar"
    elif v is None:
        ds=group.create_dataset(k, data=0)
        ds.attrs["type"]="None"
    elif nametype == 'UID':
        sUID=str(v)
        ds=group.create_dataset(k, data=sUID)
        ds.attrs["type"]=nametype
    else:
        raise TypeError( "Datatype not handled:" + nametype)


def SaveToH5py(MyDict,f,compatibility="blosc",complevel=9,group=None):
    if bCheckIfStr(f):
        fileobj=h5py.File(f, "w")
    elif type(f) is h5py._hl.files.File:
        fileobj=f
    else:
        raise TypeError( "only string or h5py.file objects are accepted for 'f' object")

    if type(MyDict) not in [dict,OrderedDict] or type(MyDict) is list:
        raise TypeError("Only dictionaries are supported to be saved")
    for (k, v )in MyDict.items():
        ProcType(k,v,fileobj,compatibility,complevel,group)
    if bCheckIfStr(f):
        fileobj.close() #if we receive a string, that means that we open and close the file

def ReadFromH5py(f,group=None,typeattr=None):
    if bCheckIfStr(f):
        fileobj=h5py.File(f, "r")
    elif type(f) is h5py._hl.files.File:
        fileobj=f
    else:
        raise TypeError( "only string or h5py.file objects are accepted for 'f' object")
    if group is None:
        group=fileobj
        MyDict={}
    else:
        if typeattr=='OrderedDict':
            MyDict=OrderedDict()
        else:
            MyDict={}
    for (namevar,val) in group.items():

        NameList=None
        typeattr=cStr(val.attrs["type"])
        if type(val) is h5py._hl.group.Group:
            ValGroup=ReadFromH5py(fileobj,val)
            if typeattr=="list" or  typeattr=="tuple" :
                NameList=namevar
            if NameList is not None:
                ListVal=[]
                for n in range(len(ValGroup)):
                    itemname = "item_%d"%n
                    ListVal.append(ValGroup[itemname])
                if typeattr=="list":
                    MyDict[NameList] = ListVal
                else:
                    MyDict[NameList] = tuple(ListVal)
            else:

                if "type_key" in val.attrs:
                    typekey= cStr(val.attrs["type_key"])
                    if typekey=='int':
                        snamevar=int(namevar)
                    elif typekey=='float':
                        snamevar=float(namevar)
                    elif typekey=='str':
                        snamevar=namevar
                    elif typekey=='unicode':
                        snamevar=namevar
                    else:
                        raise TypeError( "the type of dictionary key is not supported " + namevar + ' (' + typekey + ')')
                else:
                    snamevar=namevar
                MyDict[snamevar] = ValGroup
        elif typeattr == "None":
             MyDict[namevar]=None
        elif typeattr == "UID":
            MyDict[namevar]=UID(val[()])
        else:
            if cStr(val.attrs["type"])=="scalar":
                if type(val[()]) is numpy.int32 or type(val[()]) is numpy.int64:
                    MyDict[namevar]=int(val[()])
                else:
                    MyDict[namevar] = val[()]
            elif cStr(val.attrs["type"])=="<type 'str'>":
                MyDict[namevar] = cStr(val[()])
            else:
                MyDict[namevar] = val[()]
    if bCheckIfStr(f):
        fileobj.close()
    return MyDict



from calendar import c
from multiprocessing.dummy import Array
from unicodedata import name
import numpy as np
import os
from pathlib import Path
import platform
import time
import tempfile
from shutil import copyfile

from .StaggeredFDTD_3D_With_Relaxation_BASE import StaggeredFDTD_3D_With_Relaxation_BASE

from distutils.sysconfig import get_python_inc
from math import ceil
import ctypes
from ctypes import c_byte, c_int, c_uint32, c_float, c_wchar_p, c_uint64

mc = None

def ListDevices():
    import metalcomputebabel 
    devicesIDs=[]
    devices = metalcomputebabel.get_devices()
    for dev in devices:
        devicesIDs.append(dev.deviceName)
    return devicesIDs

class StaggeredFDTD_3D_With_Relaxation_METAL_MetalCompute(StaggeredFDTD_3D_With_Relaxation_BASE):
    '''
    This version is mainly for Mx processors and which is based in a modified+forked version of metalcompute. As X64 will be phased out, in the future the metalcompute version will take over
    '''
    def __init__(self, arguments):
        global mc
        import metalcomputebabel
        mc = metalcomputebabel
        # Definition of some constants, etc
        self.MAX_SIZE_PML = 101
        self._c_mex_type = np.zeros(12, np.uint64)
        self._c_uint_type = np.uint64(0)
        self.HOST_INDEX_MEX = np.zeros((53, 2), np.uint64)
        self.HOST_INDEX_UINT = np.zeros((3, 2), np.uint64)
        self.LENGTH_INDEX_MEX = 53
        self.LENGTH_INDEX_UINT = 3
        self.ZoneCount = arguments['SPP_ZONES']
        self._IndexDataMetal = {
            "V_x_x":0, "V_y_x":0, "V_z_x":0, "V_x_y":0, "V_y_y":0, "V_z_y":0, "V_x_z":0, "V_y_z":0, "V_z_z":0,
            "Vx":1, "Vy":1, "Vz":1,
            "Rxx":2, "Ryy":2, "Rzz":2,
            "Rxy":3, "Rxz":3, "Ryz":3,
            "Sigma_x_xx":4, "Sigma_y_xx":4, "Sigma_z_xx":4, "Sigma_x_yy":4, "Sigma_y_yy":4, "Sigma_z_yy":4, "Sigma_x_zz":4, "Sigma_y_zz":4,
            "Sigma_z_zz":5, "Sigma_x_xy":5, "Sigma_y_xy":5, "Sigma_x_xz":5, "Sigma_z_xz":5, "Sigma_y_yz":5, "Sigma_z_yz":5,
            "Sigma_xy":6, "Sigma_xz":6, "Sigma_yz":6, 
            "Sigma_xx":7, "Sigma_yy":7, "Sigma_zz":7,
            "SourceFunctions":8,
            "LambdaMiuMatOverH":9, "LambdaMatOverH":9, "MiuMatOverH":9, "TauLong":9, "OneOverTauSigma":9, "TauShear":9, "InvRhoMatH":9, "Ox":9, "Oy":9, "Oz":9, "Pressure":9, 
            "SqrAcc":10,
            "SensorOutput":11
            }
        self.C_IND = {
            "IndexSensorMap":0, "SourceMap":1, "MaterialMap": 2,
            "nStep":0, "TypeSource":1, 
            "V_x_x":0, "V_y_x":1, "V_z_x":2, "V_x_y":3, "V_y_y":4, "V_z_y":5, "V_x_z":6, "V_y_z":7, "V_z_z":8, "Vx":9, "Vy":10, "Vz":11, "Rxx":12, "Ryy":13, "Rzz":14, "Rxy":15, "Rxz":16, "Ryz":17,
            "Sigma_x_xx":18, "Sigma_y_xx":19, "Sigma_z_xx":20, "Sigma_x_yy":21, "Sigma_y_yy":22, "Sigma_z_yy":23, "Sigma_x_zz":24, "Sigma_y_zz":25, "Sigma_z_zz":26, 
            "Sigma_x_xy":27, "Sigma_y_xy":28, "Sigma_x_xz":29, "Sigma_z_xz":30, "Sigma_y_yz":31, "Sigma_z_yz":32, "Sigma_xy":33, "Sigma_xz":34, "Sigma_yz":35, "Sigma_xx":36, "Sigma_yy":37, "Sigma_zz": 38,
            "SourceFunctions":39, "LambdaMiuMatOverH":40, "LambdaMatOverH":41, "MiuMatOverH":42, "TauLong":43, "OneOverTauSigma":44, "TauShear":45, "InvRhoMatH":46, "Ox":47, "Oy":48, "Oz":49,
            "Pressure":50, "SqrAcc":51, "SensorOutput":52, 
            }
        self.LENGTH_CONST_UINT = 2
        # self.LENGTH_CONST_MEX = 1+self.MAX_SIZE_PML*4

        extra_params = {"BACKEND":"METAL"}
        
        self.bUseSingleKernel = False
        super().__init__(arguments, extra_params)
        
    def _PostInitScript(self, arguments, extra_params):
        print("Attempting Metal Initiation...")
        devices = mc.get_devices()
        SelDevice=None
        for n,dev in enumerate(devices):
            if arguments['DefaultGPUDeviceName'] in dev.deviceName:
                SelDevice=dev
                break
        if SelDevice is None:
            raise SystemError("No Metal device containing name [%s]" %(arguments['DefaultGPUDeviceName']))
        else:
            print('Selecting device: ', dev.deviceName)
        SCode = []
        SCode.append("#define mexType " + extra_params['td'] +"\n")
        SCode.append("#define METAL\n")
        if self.bUseSingleKernel:
            SCode.append("#define METAL_SINGLE_KERNEL\n")
        SCode.append("#define METALCOMPUTE\n")
        SCode.append("#define MAX_SIZE_PML 101\n")
        extra_params['SCode'] = SCode
        self.ctx = mc.Device(n)
        self.ConstantBufferUINT=np.zeros(self.LENGTH_CONST_UINT,np.uint32)
        print(self.ctx)
        if 'arm64' not in platform.platform():
            print('Setting Metal for External or AMD processor')
            self.ctx.set_external_gpu(1) 
    
    def _InitSymbol(self, IP,_NameVar,td,SCode):
        if td in ['float','double']:
            res = 'constant ' + td  + ' ' + _NameVar + ' = %0.9g;\n' %(IP[_NameVar])
        else:
            lType =' _PT '
            res = 'constant '+ lType  + _NameVar + ' = %i;\n' %(IP[_NameVar])
        SCode.append(res)
        
    def _InitSymbolArray(self, IP,_NameVar,td,SCode):
        res =  "constant "+ td + " " + _NameVar + "_pr[%i] ={\n" % (IP[_NameVar].size)
        for n in range(IP[_NameVar].size):
            if td in ['float','double']:
                res+="%.9g" % (IP[_NameVar][n])
            else:
                res+="%i" % (IP[_NameVar][n])
            if n<IP[_NameVar].size-1:
                res+=',\n'
            else:
                res+='};\n'
        SCode.append(res)   


    def _UpdateSymbol(self, IP, _NameVar,td, SCode):
        if td == "unsigned int": 
            self.constant_buffer_uint.modify(np.array(IP[_NameVar],dtype=np.uint32),int(self.C_IND[_NameVar]),1)
        else:
            raise ValueError("Something was passed incorrectly in symbol initiation.")


    def _ownGpuCalloc(self, Name,td,dims,ArraysGPUOp):
        print("Allocating for", Name, dims, "elements")
        if Name == "Snapshots":
            pass
        elif td == "float":
            self.HOST_INDEX_MEX[self.C_IND[Name]][0] = np.uint64(self._c_mex_type[self._IndexDataMetal[Name]])
            self.HOST_INDEX_MEX[self.C_IND[Name]][1] = np.uint64(dims ) 
            self._c_mex_type[self._IndexDataMetal[Name]] += dims
        elif td == "unsigned int":
            self.HOST_INDEX_UINT[self.C_IND[Name]][0] = np.uint64(self._c_uint_type)
            self.HOST_INDEX_UINT[self.C_IND[Name]][1] = np.uint64(dims)
            self._c_uint_type += dims

    
    def _CreateAndCopyFromMXVarOnGPU(self, Name,ArraysGPUOp,ArrayResCPU,flags=[]):
        print("Allocating for", Name, ArrayResCPU[Name].size, "elements")
        SizeCopy = ArrayResCPU[Name].size
        if Name in ['LambdaMiuMatOverH','LambdaMatOverH','MiuMatOverH','TauLong','OneOverTauSigma','TauShear','InvRhoMatH', 'Ox','Oy','Oz', 'SourceFunctions', 'SensorOutput','SqrAcc']: # float
            self.HOST_INDEX_MEX[self.C_IND[Name]][0] = np.uint64(self._c_mex_type[self._IndexDataMetal[Name]])
            self.HOST_INDEX_MEX[self.C_IND[Name]][1] = np.uint64(SizeCopy)
            self._c_mex_type[self._IndexDataMetal[Name]] += SizeCopy
        elif Name in ['IndexSensorMap','SourceMap','MaterialMap',]: # unsigned int
            self.HOST_INDEX_UINT[self.C_IND[Name]][0] = np.uint64(self._c_uint_type)
            self.HOST_INDEX_UINT[self.C_IND[Name]][1] = np.uint64(SizeCopy)
            self._c_uint_type += SizeCopy
    
    def _PreExecuteScript(self, arguments, ArraysGPUOp, outparams):
        print("Float entries:", np.sum(self._c_mex_type), "int entries:", self._c_uint_type)
        self._outparams = outparams
        self.mex_buffer=[]
        for nSizes in self._c_mex_type:
            self.mex_buffer.append(self.ctx.buffer(nSizes*4))
        self.uint_buffer=self.ctx.buffer(self._c_uint_type*4)
        self.constant_buffer_uint=self.ctx.buffer(self.ConstantBufferUINT)

        self._IndexManip()

        for k in ['LambdaMiuMatOverH','LambdaMatOverH','MiuMatOverH','TauLong','OneOverTauSigma','TauShear','InvRhoMatH',\
                    'Ox','Oy','Oz','SourceFunctions']:
            self._CompleteCopyToGPU(k, arguments, arguments[k].size, "float")

        for k in ['IndexSensorMap','SourceMap','MaterialMap']:
            self._CompleteCopyToGPU(k, arguments, arguments[k].size, "unsigned int")

        self.localSensor = [1, 1, 1]
        self.globalSensor = [ceil(arguments['IndexSensorMap'].size / self.localSensor[0]), 1, 1]

    def _IndexManip(self):
        index=np.zeros((self.LENGTH_INDEX_MEX,2),np.uint32)
        for i in range(self.LENGTH_INDEX_MEX):
            index[i,0] = np.uint32(np.uint64(0xFFFFFFFF) & np.uint64(self.HOST_INDEX_MEX[i][0])) # Not exactly sure if this works
            index[i,1] = np.uint32(np.uint64([self.HOST_INDEX_MEX[i][0]])>>32)
        self.index_mex=self.ctx.buffer(index)
        
        index=np.zeros((self.LENGTH_INDEX_UINT,2),np.uint32)
       
        for i in range(self.LENGTH_INDEX_UINT):
            index[i,0] = np.uint32(0xFFFFFFFF) & np.uint32(self.HOST_INDEX_UINT[i][0]) # Not exactly sure if this works
            index[i,1] = np.uint32(np.uint64([self.HOST_INDEX_MEX[i][0]])>>32)
        self.index_uint=self.ctx.buffer(index)


    def _CompleteCopyToGPU(self, Name, args, SizeCopy, td):
        if td == "float":
            self.mex_buffer[self._IndexDataMetal[Name]].modify(args[Name].flatten(order='F'),int(self.HOST_INDEX_MEX[self.C_IND[Name]][0]),int(SizeCopy))
        elif td == "unsigned int":
            self.uint_buffer.modify(args[Name].flatten(order='F'),int(self.HOST_INDEX_UINT[self.C_IND[Name]][0]),int(SizeCopy))
        else:
            raise RuntimeError("Something has gone horribly wrong.")
    

    def _InitiateCommands(self, SCode):
        AllC=''.join(SCode)
        prg = self.ctx.kernel(AllC)
        if self.bUseSingleKernel:
            PartsStress=['MAIN_1']
        else:
            PartsStress=['PML_1','PML_2','PML_3','PML_4','PML_5','PML_6','MAIN_1']
        
        self.AllStressKernels={}
        for k in PartsStress:
            self.AllStressKernels[k]=prg.function(k+"_StressKernel")

        if self.bUseSingleKernel:
            PartsParticle =['MAIN_1']
        else:
            PartsParticle=['PML_1','PML_2','PML_3','PML_4','PML_5','PML_6','MAIN_1']
        
        self.AllParticleKernels={}
        for k in PartsParticle:
            self.AllParticleKernels[k]=prg.function(k+"_ParticleKernel")
        
        self.SensorsKernel=prg.function('SensorsKernel')

    def _Execution(self, arguments, ArrayResCPU, ArrayResOP):
        TimeSteps = arguments['TimeSteps']
        InitDict = {'nStep':0, 'TypeSource':int(arguments['TypeSource'])}
        outparams=self._outparams
        DimsKernel={}
        DimsKernel['PML_1']=[outparams['PML_Thickness'],
                             outparams['N2'],
                             outparams['N3']]
        DimsKernel['PML_2']=[outparams['PML_Thickness'],
                             outparams['N2'],
                             outparams['N3']]
        DimsKernel['PML_3']=[outparams['N1']-outparams['PML_Thickness']*2,
                            outparams['PML_Thickness'],
                            outparams['N3']]
        DimsKernel['PML_4']=[outparams['N1']-outparams['PML_Thickness']*2,
                            outparams['PML_Thickness'],
                            outparams['N3']]
        DimsKernel['PML_5']=[outparams['N1']-outparams['PML_Thickness']*2,
                            outparams['N2']-outparams['PML_Thickness']*2,
                            outparams['PML_Thickness']]
        DimsKernel['PML_6']=[outparams['N1']-outparams['PML_Thickness']*2,
                            outparams['N2']-outparams['PML_Thickness']*2,
                            outparams['PML_Thickness']]
        if self.bUseSingleKernel:
            DimsKernel['MAIN_1']=[outparams['N1'],
                            outparams['N2'],
                            outparams['N3']]
            kernels=["MAIN_1"]
            
        
        else:
            DimsKernel['MAIN_1']=[outparams['N1']-outparams['PML_Thickness']*2,
                            outparams['N2']-outparams['PML_Thickness']*2,
                            outparams['N3']-outparams['PML_Thickness']*2]
            kernels =["PML_1", "PML_2", "PML_3", "PML_4", "PML_5", "PML_6", "MAIN_1"]
            

        for nStep in range(TimeSteps):
            InitDict["nStep"] = nStep
            for i in ['nStep', 'TypeSource']:
                self._UpdateSymbol(InitDict, i, 'unsigned int', [])

            self.ctx.init_command_buffer()
            AllHandles=[]
            
            for i in kernels:
                nSize=np.prod(DimsKernel[i])
                handle=self.AllStressKernels[i](nSize,self.constant_buffer_uint,
                                               self.index_mex,
                                               self.index_uint, 
                                               self.uint_buffer,
                                               self.mex_buffer[0],
                                               self.mex_buffer[1],
                                               self.mex_buffer[2],
                                               self.mex_buffer[3],
                                               self.mex_buffer[4],
                                               self.mex_buffer[5],
                                               self.mex_buffer[6],
                                               self.mex_buffer[7],
                                               self.mex_buffer[8],
                                               self.mex_buffer[9],
                                               self.mex_buffer[10],
                                               self.mex_buffer[11])
                AllHandles.append(handle)

            for i in kernels:
                nSize=np.prod(DimsKernel[i])
                handle = self.AllParticleKernels[i](nSize,self.constant_buffer_uint,
                                               self.index_mex,
                                               self.index_uint, 
                                               self.uint_buffer,
                                               self.mex_buffer[0],
                                               self.mex_buffer[1],
                                               self.mex_buffer[2],
                                               self.mex_buffer[3],
                                               self.mex_buffer[4],
                                               self.mex_buffer[5],
                                               self.mex_buffer[6],
                                               self.mex_buffer[7],
                                               self.mex_buffer[8],
                                               self.mex_buffer[9],
                                               self.mex_buffer[10],
                                               self.mex_buffer[11])

                AllHandles.append(handle)
            self.ctx.commit_command_buffer()
            self.ctx.wait_command_buffer()
            while len(AllHandles)>0:
                handle = AllHandles.pop(0) 
                del handle
            if (nStep % arguments['SensorSubSampling'])==0  and (int(nStep/arguments['SensorSubSampling'])>=arguments['SensorStart']):
                self.ctx.init_command_buffer()
                handle=self.SensorsKernel(np.prod(self.globalSensor),
                                self.constant_buffer_uint,
                                self.index_mex,
                                self.index_uint, 
                                self.uint_buffer,
                                self.mex_buffer[0],
                                self.mex_buffer[1],
                                self.mex_buffer[2],
                                self.mex_buffer[3],
                                self.mex_buffer[4],
                                self.mex_buffer[5],
                                self.mex_buffer[6],
                                self.mex_buffer[7],
                                self.mex_buffer[8],
                                self.mex_buffer[9],
                                self.mex_buffer[10],
                                self.mex_buffer[11])
                self.ctx.commit_command_buffer()   
                self.ctx.wait_command_buffer()
                del handle
        if 'arm64' not in platform.platform():
            self.ctx.sync_buffers((self.mex_buffer[0],
                                    self.mex_buffer[1],
                                    self.mex_buffer[2],
                                    self.mex_buffer[3],
                                    self.mex_buffer[4],
                                    self.mex_buffer[5],
                                    self.mex_buffer[6],
                                    self.mex_buffer[7],
                                    self.mex_buffer[8],
                                    self.mex_buffer[9],
                                    self.mex_buffer[10],
                                    self.mex_buffer[11]))
        for i in ['SqrAcc', 'SensorOutput']:
            SizeCopy = ArrayResCPU[i].size
            Shape = ArrayResCPU[i].shape
            Buffer=np.frombuffer(self.mex_buffer[self._IndexDataMetal[i]],dtype=np.float32)[int(self.HOST_INDEX_MEX[self.C_IND[i]][0]):int(self.HOST_INDEX_MEX[self.C_IND[i]][0]+SizeCopy)]
            ArrayResCPU[i][:,:,:] = np.reshape(Buffer, Shape,order='F')
     
        
        SizeBuffer = {1:0, 6:0, 7:0, 9:0}
        for i in ['Vx', 'Vy', 'Vz', 'Sigma_xx', 'Sigma_yy', 'Sigma_zz', 'Sigma_xy', 'Sigma_xz', 'Sigma_yz']:
            SizeBuffer[self._IndexDataMetal[i]] += ArrayResCPU[i].size* self.ZoneCount
        
        for i in ["LambdaMiuMatOverH", "LambdaMatOverH", "MiuMatOverH", "TauLong", "OneOverTauSigma", "TauShear", "InvRhoMatH", "Ox", "Oy", "Oz"]:
            SizeBuffer[9] += arguments[i].size
        
        SizeBuffer[9] += ArrayResCPU['Pressure'].size* self.ZoneCount

        for i in ['Vx', 'Vy', 'Vz', 'Sigma_xx', 'Sigma_yy', 'Sigma_zz', 'Sigma_xy', 'Sigma_xz', 'Sigma_yz', 'Pressure']:
            SizeCopy = ArrayResCPU[i].size * self.ZoneCount
            sz=ArrayResCPU[i].shape
            Shape = (sz[0],sz[1],sz[2],self.ZoneCount)
            Buffer=np.frombuffer(self.mex_buffer[self._IndexDataMetal[i]],dtype=np.float32)[int(self.HOST_INDEX_MEX[self.C_IND[i]][0]):int(self.HOST_INDEX_MEX[self.C_IND[i]][0]+SizeCopy)]
            Buffer=Buffer.reshape(Shape,order='F')
            ArrayResCPU[i][:,:,:] = np.sum(Buffer,axis=3)/self.ZoneCount
          
        del self.constant_buffer_uint
        del self.index_mex
        del self.index_uint 
        del self.uint_buffer
        while len(self.mex_buffer)>0:
            handle = self.mex_buffer.pop()
            del handle

def StaggeredFDTD_3D_METAL(arguments):
    Instance = StaggeredFDTD_3D_With_Relaxation_METAL_MetalCompute(arguments)
    Results = Instance.Results
    return Results



import numpy as np
import _FDTDStaggered3D_with_relaxation_double as FDTD_double;
import _FDTDStaggered3D_with_relaxation_single as FDTD_single;
import time


def StaggeredFDTD_3D(arguments):
    if (type(arguments)!=dict):
        raise TypeError( "The input parameter must be a dictionary")

    for key in arguments.keys():
        if type(arguments[key])==np.ndarray:
            if np.isfortran(arguments[key])==False:
                #print "StaggeredFDTD_3D: Converting ndarray " + key + " to Fortran convention";
                arguments[key] = np.asfortranarray(arguments[key]);
        elif type(arguments[key])!=str:
            arguments[key]=np.array((arguments[key]))
    t0=time.time()
    arguments['PI_OCL_PATH']='' #unused in OpenMP but needed in the low level function for completeness
    arguments['kernelfile']='' #these are unused in OpenMP but passed for completeness
    arguments['kernbinfile']='' #these are unused in OpenMP but passed for completeness
    
    if arguments['DT'].dtype==np.dtype('float32'):
        Results= FDTD_single.FDTDStaggered_3D(arguments)
    else:
        Results = FDTD_double.FDTDStaggered_3D(arguments)
    t0=time.time()-t0
    print ('Time to run low level FDTDStaggered_3D =', t0)
    return Results



from multiprocessing.sharedctypes import Value
import numpy as np
import os
from pathlib import Path
import platform
import time
import tempfile
from shutil import copyfile

#we will generate the _kernel-opencl.c file when importing
from distutils.sysconfig import get_python_inc

TotalAllocs=0

MASKID={}
MASKID['Vx']=           0x0000000001
MASKID['Vy']  =         0x0000000002
MASKID['Sigmaxx']  =    0x0000000004
MASKID['Sigmayy'] =     0x0000000008
MASKID['Sigmaxy'] =     0x0000000010
MASKID['Pressure'] =    0x0000000020
MASKID['Pressure_gx'] = 0x0000000040
MASKID['Pressure_gy'] = 0x0000000080
MASKID['SEL_RMS'] =0x0000000001
MASKID['SEL_PEAK']=0x0000000002
class StaggeredFDTD_2D_With_Relaxation_BASE():
    def __init__(self, arguments, extra_params={}):
        global NumberSelRMSPeakMaps # Is it necessary to keep these global anymore?
        global NumberSelSensorMaps
        global TotalAllocs
        
        IncludeDir=str(Path(__file__).parent.absolute())+os.sep


        if (type(arguments)!=dict):
            raise TypeError( "The input parameter must be a dictionary")

        for key in arguments.keys():
            if type(arguments[key])==np.ndarray:
                if np.isfortran(arguments[key])==False:
                    #print "StaggeredFDTD_3D: Converting ndarray " + key + " to Fortran convention";
                    arguments[key] = np.asfortranarray(arguments[key])
            elif type(arguments[key])!=str:
                arguments[key]=np.array((arguments[key]))

        if arguments['DT'].dtype==np.dtype('float32'):
            dtype=np.float32
        elif extra_params["BACKEND"] == "METAL":
            raise SystemError("Metal backend only supports single precision")
        else:
            dtype=np.float64

        gpu_kernelSrc=IncludeDir+'_gpu_kernel2D.c'
        index_src=IncludeDir+'_indexing2D.h'


        td = 'float'
        if dtype is np.float64:
            td='double'
        extra_params['td'] = td
    
        t0=time.time()
        
        NumberSelRMSPeakMaps=0
        NumberSelSensorMaps=0
        TotalAllocs=0
        
        outparams = self._PrepParamsForKernel(arguments)
        
        #we prepare the kernel code

        self._PostInitScript(arguments, extra_params)
        
        
        SCode = extra_params["SCode"]
        with open(index_src) as f:
            SCode+=f.readlines()
        
    
        LParamFloat = ['DT']
        LParamInt=["N1","N2", "Limit_I_low_PML", "Limit_J_low_PML", "Limit_I_up_PML","Limit_J_up_PML",\
                 "SizeCorrI","SizeCorrJ","PML_Thickness","NumberSources", "LengthSource","ZoneCount",\
                "SizePMLxp1","SizePMLyp1","SizePML","SizePMLxp1yp1","NumberSensors","TimeSteps","SelRMSorPeak",\
                "SelMapsRMSPeak","IndexRMSPeak_Vx","IndexRMSPeak_Vy", "IndexRMSPeak_Sigmaxx",\
                "IndexRMSPeak_Sigmayy","IndexRMSPeak_Sigmaxy",\
                "IndexRMSPeak_Pressure","IndexRMSPeak_Pressure_gx","IndexRMSPeak_Pressure_gy","NumberSelRMSPeakMaps","SelMapsSensors","IndexSensor_Vx","IndexSensor_Vy",\
                "IndexSensor_Sigmaxx","IndexSensor_Sigmayy","IndexSensor_Sigmaxy",\
                "IndexSensor_Pressure","IndexSensor_Pressure_gx","IndexSensor_Pressure_gy","NumberSelSensorMaps","SensorSubSampling",
                "SensorStart"]
        LParamArray=['InvDXDTplus','DXDTminus','InvDXDTplushp','DXDTminushp']
        tt =LParamFloat+LParamInt+LParamArray
        for k in tt:
            if k not in outparams:
                print(k)
        assert len(outparams)==(len(LParamFloat)+len(LParamInt)+len(LParamArray))
        for k in LParamFloat:
            self._InitSymbol(outparams,k,td,SCode)
        for k in LParamInt:
            self._InitSymbol(outparams,k,'unsigned int',SCode)
        for k in LParamArray:
            self._InitSymbolArray(outparams,k,td,SCode)

        with open(gpu_kernelSrc) as f:
            SCode+=f.readlines()
        AllC=''
        for l in SCode:
            AllC+=l
    
        N1=arguments['N1']
        N2=arguments['N2']

                    
        ArrayResCPU={}
        for k in ['Sigma_xx','Sigma_yy','Pressure']:
            ArrayResCPU[k]=np.zeros((N1,N2),dtype,order='F')
        for k in ['Sigma_xy']:
            ArrayResCPU[k]=np.zeros((N1+1,N2+1),dtype,order='F')
        ArrayResCPU['Vx']=np.zeros((N1+1,N2),dtype,order='F')
        ArrayResCPU['Vy']=np.zeros((N1,N2+1),dtype,order='F')
    
        if 	(arguments['SelRMSorPeak'] &  MASKID['SEL_PEAK']) and (arguments['SelRMSorPeak'] &  MASKID['SEL_RMS']):
            #both peak and RMS
            updims=2 
        else:
            updims=1

        ArrayResCPU['SqrAcc']=np.zeros((N1,N2,outparams['NumberSelRMSPeakMaps'],updims),dtype,order='F')
        Ns=1
        NumberSnapshots=arguments['SnapshotsPos'].size
        NumberSensors=arguments['IndexSensorMap'].size
        if NumberSnapshots>0:
            Ns=NumberSnapshots
        ArrayResCPU['Snapshots']=np.zeros((N1,N2,Ns),dtype,order='F')
        TimeSteps=arguments['TimeSteps']
        SensorSubSampling=arguments['SensorSubSampling']
        SensorStart=arguments['SensorStart']
        print("Number Selected Sensor Maps:", outparams['NumberSelRMSPeakMaps'])
        ArrayResCPU['SensorOutput']=np.zeros((NumberSensors,int(TimeSteps/SensorSubSampling)+1-SensorStart,outparams['NumberSelSensorMaps']),dtype,order='F')
        
        self._InitiateCommands(AllC)

        ArraysGPUOp={}
        if extra_params["BACKEND"] in ["OPENCL","CUDA"]:
            for k in ['LambdaMiuMatOverH','LambdaMatOverH','MiuMatOverH','TauLong','OneOverTauSigma','TauShear','InvRhoMatH',\
                        'Ox','Oy','SourceFunctions','IndexSensorMap','SourceMap','MaterialMap']:            
                self._CreateAndCopyFromMXVarOnGPU(k,ArraysGPUOp,arguments)
        for k in ['V_x_x','V_y_x']:
            self._ownGpuCalloc(k,td,outparams['SizePMLxp1']*outparams['ZoneCount'],ArraysGPUOp)
        for k in ['V_x_y','V_y_y']:
            self._ownGpuCalloc(k,td,outparams['SizePMLyp1']*outparams['ZoneCount'],ArraysGPUOp)
        for k in ['Sigma_x_xx','Sigma_y_xx','Sigma_x_yy','Sigma_y_yy']:
            self._ownGpuCalloc(k,td,outparams['SizePML']*outparams['ZoneCount'],ArraysGPUOp)
        for k in ['Sigma_x_xy','Sigma_y_xy']:
            self._ownGpuCalloc(k,td,outparams['SizePMLxp1yp1']*outparams['ZoneCount'],ArraysGPUOp)
        for k in ['Rxx','Ryy']:
            self._ownGpuCalloc(k,td,ArrayResCPU['Sigma_xx'].size*outparams['ZoneCount'],ArraysGPUOp)
        for k in ['Rxy']:
            self._ownGpuCalloc(k,td,ArrayResCPU['Sigma_xy'].size*outparams['ZoneCount'],ArraysGPUOp)
        if extra_params['BACKEND'] in ['OPENCL','CUDA']:
            for k in ['Vx','Vy','Sigma_xx','Sigma_yy','Pressure','Sigma_xy','Snapshots']:
                self._ownGpuCalloc(k,td,ArrayResCPU[k].size*outparams['ZoneCount'],ArraysGPUOp)
            for k in ['SensorOutput','SqrAcc']:
                self._CreateAndCopyFromMXVarOnGPU(k, ArraysGPUOp, ArrayResCPU)
            
        else: # ORDER DOES MATTER FOR METAL, AS IT INVOLVES MANIPULATING AND READING _c_uint_type OR _c_mex_type******
            for k in ['LambdaMiuMatOverH','LambdaMatOverH','MiuMatOverH','TauLong','OneOverTauSigma','TauShear','InvRhoMatH',\
                        'Ox','Oy','SourceFunctions','IndexSensorMap','SourceMap','MaterialMap']:            
                self._CreateAndCopyFromMXVarOnGPU(k,ArraysGPUOp,arguments)
            
            for k in ['Vx','Vy','Sigma_xx','Sigma_yy','Sigma_xy','Pressure','Snapshots']:
                self._ownGpuCalloc(k,td,ArrayResCPU[k].size*outparams['ZoneCount'],ArraysGPUOp)

            for k in ['SqrAcc','SensorOutput']:
                self._CreateAndCopyFromMXVarOnGPU(k, ArraysGPUOp, ArrayResCPU)

        self._PreExecuteScript(arguments, ArraysGPUOp, outparams)

        self._Execution(arguments, ArrayResCPU, ArraysGPUOp)
            
        t0=time.time()-t0
        print ('Time to run low level FDTDStaggered_3D =', t0)

        AllC=''
        
        self.CreateResults(ArrayResCPU)

        return

    def _PostInitScript(self, arguments):
        raise NotImplementedError("This block must be implemented in a child class!")

    def _InitSymbol(self, IP,_NameVar,td,SCode=[]):
        raise NotImplementedError("This block must be implemented in a child class!")
    
    def _InitSymbolArray(self, IP,_NameVar,td,SCode=[]):
        raise NotImplementedError("This block must be implemented in a child class!")
    
    def _ownGpuCalloc(self, Name,td,dims,ArraysGPUOp,flags):
        raise NotImplementedError("This block must be implemented in a child class")
    
    def _CreateAndCopyFromMXVarOnGPU(self, Name,ctx,ArraysGPUOp,ArrayResCPU,flags):
        raise NotImplementedError("This block must be implemented in a child class")
    
    def _PrepParamsForKernel(self, arguments):
        global NumberSelRMSPeakMaps
        global NumberSelSensorMaps


        copyParams=['DT','N1','N2','SensorSubSampling','SensorStart','LengthSource','TimeSteps',\
                'SelRMSorPeak','SelMapsRMSPeak','SelMapsSensors','InvDXDTplus','DXDTminus','InvDXDTplushp','DXDTminushp']
        outparams={}
        for c in copyParams:
            outparams[c]=arguments[c]
            
        outparams['PML_Thickness']=arguments['PMLThickness']
            
        outparams['NumberSources']=arguments['SourceFunctions'].shape[0]
        outparams['ZoneCount']=arguments['SPP_ZONES']
        outparams['NumberSensors']=arguments['IndexSensorMap'].size
        
        
        outparams['Limit_I_low_PML']=outparams['PML_Thickness']-1
        outparams['Limit_I_up_PML']=arguments['N1']-outparams['PML_Thickness']
        outparams['Limit_J_low_PML']=outparams['PML_Thickness']-1
        outparams['Limit_J_up_PML']=arguments['N2']-outparams['PML_Thickness']

        outparams['SizeCorrI']=arguments['N1']-2*outparams['PML_Thickness']
        outparams['SizeCorrJ']=arguments['N2']-2*outparams['PML_Thickness']

        #//The size of the matrices where the PML is valid depends on the size of the PML barrier
        outparams['SizePML']= arguments['N1']*arguments['N2'] - outparams['SizeCorrI']*outparams['SizeCorrJ']+1
        outparams['SizePMLxp1']= (arguments['N1']+1)*(arguments['N2']) - outparams['SizeCorrI']*outparams['SizeCorrJ']+1
        outparams['SizePMLyp1']= arguments['N1']*(arguments['N2']+1) - outparams['SizeCorrI']*outparams['SizeCorrJ']+1
        outparams['SizePMLxp1yp1']= (arguments['N1']+1)*(arguments['N2']+1) - outparams['SizeCorrI']*outparams['SizeCorrJ']+1

        for k in ['Vx','Vy','Sigmaxx','Sigmayy','Sigmaxy','Pressure','Pressure_gx','Pressure_gy']:
            outparams['IndexRMSPeak_'+k]=0
            outparams['IndexSensor_'+k]=0
            if arguments['SelMapsRMSPeak'] & MASKID[k]:
                outparams['IndexRMSPeak_'+k]=NumberSelRMSPeakMaps
                NumberSelRMSPeakMaps+=1
            if arguments['SelMapsSensors'] & MASKID[k]:
                outparams['IndexSensor_'+k]=NumberSelSensorMaps
                NumberSelSensorMaps+=1
        outparams['NumberSelRMSPeakMaps']=NumberSelRMSPeakMaps
        outparams['NumberSelSensorMaps']=NumberSelSensorMaps
        return outparams

    def _Execution(self, arguments):
        raise NotImplementedError("This block must be implemented in a child class")
    
    def CreateResults(self, ArrayResCPU):
        self.Results = ArrayResCPU['SensorOutput'],\
                {'Vx':ArrayResCPU['Vx'],\
                'Vy':ArrayResCPU['Vy'],\
                'Sigma_xx':ArrayResCPU['Sigma_xx'],\
                'Sigma_yy':ArrayResCPU['Sigma_yy'],\
                'Sigma_xy':ArrayResCPU['Sigma_xy'],\
                'Pressure':ArrayResCPU['Pressure']},\
                ArrayResCPU['SqrAcc'],ArrayResCPU['Snapshots'] 


import numpy as np
import cupy as cp

import time

from .StaggeredFDTD_2D_With_Relaxation_BASE import StaggeredFDTD_2D_With_Relaxation_BASE

import struct
TotalAllocs=0

def ListDevices():
    devCount = cp.cuda.runtime.getDeviceCount()
    devicesIDs=[]
    for deviceID in range(0, devCount):
        d=cp.cuda.runtime.getDeviceProperties(deviceID)
        devicesIDs.append(d['name'].decode('UTF-8'))
    return devicesIDs

class StaggeredFDTD_2D_With_Relaxation_CUDA(StaggeredFDTD_2D_With_Relaxation_BASE):
    def __init__(self, arguments):
        extra_params = {"BACKEND":"CUDA"}
        super().__init__(arguments, extra_params)

    def _PostInitScript(self, arguments, extra_params):
        global TotalAllocs
        global _bCudaInit
        TotalAllocs=0
        devCount = cp.cuda.runtime.getDeviceCount()
        print("Number of CUDA devices found:", devCount)
        if devCount == 0:
            raise SystemError("There are no CUDA devices.")
        
        devicelist = {}

        for deviceID in range(0, devCount):
            d=cp.cuda.runtime.getDeviceProperties(deviceID)
            print("Found device", str(deviceID + 1) + "/" + str(devCount),d['name'].decode('UTF-8'))
            device = cp.cuda.Device(deviceID)
            if arguments['DefaultGPUDeviceName'] in d['name'].decode('UTF-8'):
                devicelist[deviceID] = device
    
        if len(devicelist) == 0:
            raise SystemError("There are no devices supporting CUDA or that matches selected device.")
        elif len(devicelist) < arguments['DefaultGPUDeviceNumber']:
            print("The requested device,", arguments['DefaultGPUDeviceNumber'], "(0-base index) is more than the number of devices available", len(devicelist))
            raise IndexError("Unable to select requested device.")

        print("Selecting device", arguments['DefaultGPUDeviceName'], "with number: ", arguments['DefaultGPUDeviceNumber'])

        self.device = devicelist[int(arguments['DefaultGPUDeviceNumber'])]
        self.context = cp.cuda.Device(self.device)
        print("Context Created!")
        self.context.use()
        SCode=[]
        SCode.append("#define mexType " + extra_params['td'] +"\n")
        SCode.append("#define CUDA\n")
        extra_params['SCode'] = SCode
        

    def _InitiateCommands(self, AllC):
        self._prgcuda  = cp.RawModule(code=AllC)

        PartsStress=['MAIN_1']
        self.AllStressKernels={}
        for k in PartsStress:
            self.AllStressKernels[k]=self._prgcuda.get_function(k+"_StressKernel")

        PartsParticle=['MAIN_1']
        self.AllParticleKernels={}
        for k in PartsParticle:
            self.AllParticleKernels[k]=self._prgcuda.get_function(k+"_ParticleKernel")
        
        self.SensorsKernel=self._prgcuda.get_function("SensorsKernel")

    def _InitSymbol(self, IP,_NameVar,td,SCode=[]):

        if td in ['float','double']:
            res = '__constant__ ' + td  + ' ' + _NameVar + ' = %0.9g;\n' %(IP[_NameVar])
        else:
            lType =' _PT '
            res = '__constant__ '+ lType  + _NameVar + ' = %i;\n' %(IP[_NameVar])
        SCode.append(res)    
        
    def _InitSymbolArray(self, IP,_NameVar,td,SCode=[]):
        res =  "__constant__ "+ td + " gpu" + _NameVar + "pr[%i] ={\n" % (IP[_NameVar].size)
        for n in range(IP[_NameVar].size):
            if td in ['float','double']:
                res+="%.9g" % (IP[_NameVar][n])
            else:
                res+="%i" % (IP[_NameVar][n])
            if n<IP[_NameVar].size-1:
                res+=',\n'
            else:
                res+='};\n'
        SCode.append(res)   

    def _ownGpuCalloc(self, Name,td,dims,ArraysGPUOp,flags=None):
        global TotalAllocs

        if td == 'float':
            dt=cp.float32
        elif td =='unsigned int':
            dt=cp.uint32
        else: # double
            dt=cp.float64
        print('Allocating for',Name,dims,'elements')
        ArraysGPUOp[Name]=cp.zeros((dims),dt)
        TotalAllocs+=1   
    
    def _CreateAndCopyFromMXVarOnGPU(self, Name,ArraysGPUOp,ArrayResCPU,flags=None):
        global TotalAllocs
        print('Allocating for',Name,ArrayResCPU[Name].size,'elements')
        ArraysGPUOp[Name]=cp.asarray(ArrayResCPU[Name])
        TotalAllocs+=1

    def _Execution(self, arguments, ArrayResCPU, ArraysGPUOp):
        TimeSteps=arguments['TimeSteps']
        NumberSensors=arguments['IndexSensorMap'].size

        for nStep in range(TimeSteps):
            for AllKrnl in [self.AllStressKernels,self.AllParticleKernels]:
                for k in AllKrnl:
                    AllKrnl[k](self.AllGridSizes[k],
                              self.AllBlockSizes[k],
                        (ArraysGPUOp["V_x_x"],
                        ArraysGPUOp["V_y_x"],
                        ArraysGPUOp["V_x_y"],
                        ArraysGPUOp["V_y_y"],
                        ArraysGPUOp["Vx"],
                        ArraysGPUOp["Vy"],
                        ArraysGPUOp["Rxx"],
                        ArraysGPUOp["Ryy"],
                        ArraysGPUOp["Rxy"],
                        ArraysGPUOp["Sigma_x_xx"],
                        ArraysGPUOp["Sigma_y_xx"],
                        ArraysGPUOp["Sigma_x_yy"],
                        ArraysGPUOp["Sigma_y_yy"],
                        ArraysGPUOp["Sigma_x_xy"],
                        ArraysGPUOp["Sigma_y_xy"],
                        ArraysGPUOp["Sigma_xy"],
                        ArraysGPUOp["Sigma_xx"],
                        ArraysGPUOp["Sigma_yy"],
                        ArraysGPUOp["SourceFunctions"],
                        ArraysGPUOp["LambdaMiuMatOverH"],
                        ArraysGPUOp["LambdaMatOverH"],
                        ArraysGPUOp["MiuMatOverH"],
                        ArraysGPUOp["TauLong"],
                        ArraysGPUOp["OneOverTauSigma"],
                        ArraysGPUOp["TauShear"],
                        ArraysGPUOp["InvRhoMatH"],
                        ArraysGPUOp["SqrAcc"],
                        ArraysGPUOp["MaterialMap"],
                        ArraysGPUOp["SourceMap"],
                        ArraysGPUOp["Ox"],
                        ArraysGPUOp["Oy"],
                        ArraysGPUOp["Pressure"],
                        nStep,
                        arguments['TypeSource'])
                        )
                    cp.cuda.Stream.null.synchronize() 

            if (nStep % arguments['SensorSubSampling'])==0  and (int(nStep/arguments['SensorSubSampling'])>=arguments['SensorStart']):
                self.SensorsKernel(self.GridSensors,
                        self.BlockSensors,
                        (ArraysGPUOp["V_x_x"],
                        ArraysGPUOp["V_y_x"],
                        ArraysGPUOp["V_x_y"],
                        ArraysGPUOp["V_y_y"],
                        ArraysGPUOp["Vx"],
                        ArraysGPUOp["Vy"],
                        ArraysGPUOp["Rxx"],
                        ArraysGPUOp["Ryy"],
                        ArraysGPUOp["Rxy"],
                        ArraysGPUOp["Sigma_x_xx"],
                        ArraysGPUOp["Sigma_y_xx"],
                        ArraysGPUOp["Sigma_x_yy"],
                        ArraysGPUOp["Sigma_y_yy"],
                        ArraysGPUOp["Sigma_x_xy"],
                        ArraysGPUOp["Sigma_y_xy"],
                        ArraysGPUOp["Sigma_xy"],
                        ArraysGPUOp["Sigma_xx"],
                        ArraysGPUOp["Sigma_yy"],
                        ArraysGPUOp["SourceFunctions"],
                        ArraysGPUOp["LambdaMiuMatOverH"],
                        ArraysGPUOp["LambdaMatOverH"],
                        ArraysGPUOp["MiuMatOverH"],
                        ArraysGPUOp["TauLong"],
                        ArraysGPUOp["OneOverTauSigma"],
                        ArraysGPUOp["TauShear"],
                        ArraysGPUOp["InvRhoMatH"],
                        ArraysGPUOp["SqrAcc"],
                        ArraysGPUOp["MaterialMap"],
                        ArraysGPUOp["SourceMap"],
                        ArraysGPUOp["Ox"],
                        ArraysGPUOp["Oy"],
                        ArraysGPUOp["Pressure"],
                        ArraysGPUOp['SensorOutput'],
                        ArraysGPUOp['IndexSensorMap'],
                        nStep))
                cp.cuda.Stream.null.synchronize()

        bFirstCopy=True
        events=[]
        for k in ['Vx','Vy','Sigma_xx','Sigma_yy',
            'Sigma_xy','Pressure']:
            sz=ArrayResCPU[k].shape
            tempArray=ArraysGPUOp[k].get().reshape((sz[0],sz[1],arguments['SPP_ZONES']), order='F')
            ArrayResCPU[k][:,:]=tempArray.sum(axis=2)/arguments['SPP_ZONES']

        for k in ['SqrAcc','Snapshots','SensorOutput']:
            ArrayResCPU[k]=ArraysGPUOp[k].get()
            
        cp.cuda.Stream.null.synchronize()
        
        for k in list(ArraysGPUOp.keys()):
            h= ArraysGPUOp.pop(k)
            del h


    def _PreExecuteScript(self, arguments, ArraysGPUOp, dummy):
        self.TimeSteps = arguments['TimeSteps']
        N1=arguments['N1']
        N2=arguments['N2']
        
        if arguments['ManualLocalSize'][0]!=-1:
            AllBlockSizes=(arguments['ManualLocalSize'][0],
                           arguments['ManualLocalSize'][1])
        else:
            AllBlockSizes=(8,2)

        self.AllBlockSizes={}
        self.AllBlockSizes['MAIN_1']=AllBlockSizes
        self.AllGridSizes={}

        if arguments['ManualGroupSize'][0]!=-1:
            self.AllGridSizes['MAIN_1']=(arguments['ManualGroupSize'][0],
                            arguments['ManualGroupSize'][1])
        else:
            self.AllGridSizes['MAIN_1']=(int(N1//self.AllBlockSizes['MAIN_1'][0]+1),
                              int(N2//self.AllBlockSizes['MAIN_1'][1]+1))

        self.BlockSensors=(64,1)
        self.GridSensors=(int(arguments['IndexSensorMap'].size//self.BlockSensors[0]+1),1)


def StaggeredFDTD_2D_CUDA(arguments):
    Instance = StaggeredFDTD_2D_With_Relaxation_CUDA(arguments)
    return Instance.Results

